@@ -1,6 +1,6 @@
 /* stringhelp.c -  standard string helper functions
  * Copyright (C) 1998, 1999, 2000, 2001, 2003, 2004, 2005,
- *               2006  Free Software Foundation, Inc.
+ *               2006, 2007  Free Software Foundation, Inc.
  *
  * This file is part of JNLIB.
  *
@@ -30,12 +30,12 @@
 #endif
 
 #include "libjnlib-config.h"
-#ifdef JNLIB_NEED_UTF8CONV
 #include "utf8conv.h"
-#endif
 #include "stringhelp.h"
 
 
+#define tohex_lower(n) ((n) < 10 ? ((n) + '0') : (((n) - 10) + 'a'))
+
 /*
  * Look for the substring SUB in buffer and return a pointer to that
  * substring in BUFFER or NULL if not found.
@@ -241,8 +241,11 @@ make_basename(const char *filepath, const char *inputpath)
 {
     char *p;
 
+
 #ifdef __riscos__
     return riscos_make_basename(filepath, inputpath);
+#else
+    (void)inputpath;
 #endif
 
     if ( !(p=strrchr(filepath, '/')) )
@@ -273,10 +276,10 @@ make_dirname(const char *filepath)
     char *p;
 
     if ( !(p=strrchr(filepath, '/')) )
-      #ifdef HAVE_DRIVE_LETTERS
+#ifdef HAVE_DRIVE_LETTERS
 	if ( !(p=strrchr(filepath, '\\')) )
 	    if ( !(p=strrchr(filepath, ':')) )
-      #endif
+#endif
 	      {
 		return jnlib_xstrdup(".");
 	      }
@@ -444,7 +447,6 @@ print_sanitized_buffer (FILE *fp, const void *buffer, size_t length,
 }
 
 
-#ifdef JNLIB_NEED_UTF8CONV
 size_t 
 print_sanitized_utf8_buffer (FILE *fp, const void *buffer,
                              size_t length, int delim)
@@ -470,8 +472,6 @@ print_sanitized_utf8_buffer (FILE *fp, const void *buffer,
   else
     return print_sanitized_buffer (fp, p, length, delim);
 }
-#endif /*JNLIB_NEED_UTF8CONV*/
-
 
 
 size_t 
@@ -487,7 +487,6 @@ print_sanitized_string (FILE *fp, const char *string, int delim)
   return string? print_sanitized_buffer (fp, string, strlen (string), delim):0;
 }
 
-#ifdef JNLIB_NEED_UTF8CONV
 size_t 
 print_sanitized_utf8_string (FILE *fp, const char *string, int delim)
 {
@@ -495,7 +494,6 @@ print_sanitized_utf8_string (FILE *fp, const char *string, int delim)
                                               string, strlen (string),
                                               delim) : 0;
 }
-#endif /*JNLIB_NEED_UTF8CONV*/
 
 /* Create a string from the buffer P_ARG of length N which is suitable for
    printing.  Caller must release the created string using xfree. */
@@ -554,6 +552,23 @@ sanitize_buffer (const void *p_arg, size_t n, int delim)
 }
 
 
+/* Given a string containing an UTF-8 encoded text, return the number
+   of characters in this string.  It differs from strlen in that it
+   only counts complete UTF-8 characters.  Note, that this function
+   does not take combined characters into account.  */
+size_t
+utf8_charcount (const char *s)
+{
+  size_t n;
+
+  for (n=0; *s; s++)
+    if ( (*s&0xc0) != 0x80 ) /* Exclude continuation bytes: 10xxxxxx */
+      n++;
+
+  return n;
+}
+
+
 /****************************************************
  **********  W32 specific functions  ****************
  ****************************************************/
@@ -811,7 +826,53 @@ memrchr (const void *buffer, int c, size_t n)
 
   for (p += n; n ; n--)
     if (*--p == c)
-      return p;
+      return (void *)p;
   return NULL;
 }
 #endif /*HAVE_MEMRCHR*/
+
+
+/* Percent-escape the string STR by replacing colons with '%3a'.  If
+   EXTRA is not NULL all characters in it are also escaped. */
+char *
+percent_escape (const char *str, const char *extra)
+{
+  int i, j;
+  char *ptr;
+
+  if (!str)
+    return NULL;
+
+  for (i=j=0; str[i]; i++)
+    if (str[i] == ':' || str[i] == '%' || (extra && strchr (extra, str[i])))
+      j++;
+  ptr = jnlib_xmalloc (i + 2 * j + 1);
+  i = 0;
+  while (*str)
+    {
+      if (*str == ':')
+	{
+	  ptr[i++] = '%';
+	  ptr[i++] = '3';
+	  ptr[i++] = 'a';
+	}
+      else if (*str == '%')
+	{
+	  ptr[i++] = '%';
+	  ptr[i++] = '2';
+	  ptr[i++] = '5';
+	}
+      else if (extra && strchr (extra, *str))
+        {
+	  ptr[i++] = '%';
+          ptr[i++] = tohex_lower ((*str>>4)&15);
+          ptr[i++] = tohex_lower (*str&15);
+        }
+      else
+	ptr[i++] = *str;
+      str++;
+    }
+  ptr[i] = '\0';
+
+  return ptr;
+}
