@@ -1,6 +1,6 @@
 /* crlcache.c - LDAP access
- *      Copyright (C) 2002 Klarälvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004, 2005 g10 Code GmbH
+ * Copyright (C) 2002 Klarälvdalens Datakonsult AB
+ * Copyright (C) 2003, 2004, 2005, 2008 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -15,8 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 /* 
@@ -94,7 +93,13 @@
 #include <dirent.h>
 #include <fcntl.h>
 #include <unistd.h>
+#ifndef HAVE_W32_SYSTEM
 #include <sys/utsname.h>
+#endif
+#ifdef MKDIR_TAKES_ONE_ARG
+#undef mkdir
+#define mkdir(a,b) mkdir(a)
+#endif
 
 #include "dirmngr.h"
 #include "validate.h"
@@ -117,6 +122,7 @@
 
 
 static const char oidstr_crlNumber[] = "2.5.29.20";
+static const char oidstr_issuingDistributionPoint[] = "2.5.29.28";
 static const char oidstr_authorityKeyIdentifier[] = "2.5.29.35";
 
 
@@ -732,12 +738,19 @@ update_dir (crl_cache_t cache)
   {
     char *tmpbuf, *p;
     const char *nodename;
+#ifndef HAVE_W32_SYSTEM
     struct utsname utsbuf;
+#endif
     
+#ifdef HAVE_W32_SYSTEM
+    nodename = "unknown";
+#else
     if (uname (&utsbuf))
       nodename = "unknown";
     else
       nodename = utsbuf.nodename;
+#endif
+
     tmpbuf = xmalloc (strlen (nodename) + 100);
     sprintf (tmpbuf, "DIR-tmp-%s-%u-%p.txt.tmp",
              nodename, (unsigned int)getpid (), tmpbuf);
@@ -847,6 +860,10 @@ update_dir (crl_cache_t cache)
     }
   fpout = NULL;
 
+#ifdef HAVE_W32_SYSTEM
+  /* No atomic mv on W32 systems.  */
+  unlink (fname);
+#endif
   if (rename (tmpfname, fname))
     {
       err = gpg_error_from_errno (errno);
@@ -1183,6 +1200,8 @@ cache_isvalid (ctrl_t ctrl, const char *issuer_hash,
   dirmngr_isotime_t current_time;
   size_t n;
 
+  (void)ctrl;
+
   entry = find_entry (cache->entries, issuer_hash);
   if (!entry)
     {
@@ -1426,7 +1445,7 @@ start_sig_check (ksba_crl_t crl, gcry_md_hd_t *md, int *algo)
       return err;
     }
   if (DBG_HASHING)
-    gcry_md_start_debug (*md, "crl");
+    gcry_md_debug (*md, "crl");
 
   ksba_crl_set_hash_function (crl, HASH_FNC, *md);  
   return 0;
@@ -1437,7 +1456,7 @@ start_sig_check (ksba_crl_t crl, gcry_md_hd_t *md, int *algo)
    should return 0 on a good signature, GPG_ERR_BAD_SIGNATURE if the
    signature does not verify or any other error code. CRL is the CRL
    object we are working on, MD the hash context and ISSUER_CERT the
-   certificate of the CRL issuer. */
+   certificate of the CRL issuer.  This function closes MD.  */
 static gpg_error_t
 finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
                   ksba_cert_t issuer_cert)
@@ -1450,6 +1469,7 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
   gcry_sexp_t s_sig = NULL, s_hash = NULL, s_pkey = NULL;
   int i;
 
+  /* This also stops debugging on the MD.  */
   gcry_md_final (md);
 
   /* Get and convert the signature value. */
@@ -1513,11 +1533,22 @@ finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
   gcry_sexp_release (s_sig);
   gcry_sexp_release (s_hash);
   gcry_sexp_release (s_pkey);
-  
+  gcry_md_close (md);
+
   return err;
 }
 
 
+/* Call this to match a start_sig_check that can not be completed
+   normally.  */
+static void
+abort_sig_check (ksba_crl_t crl, gcry_md_hd_t md)
+{
+  (void)crl;
+  gcry_md_close (md);
+}
+
+
 /* Workhorse of the CRL loading machinery.  The CRL is read using the
    CRL object and stored in the data base file DB with the name FNAME
    (only used for printing error messages).  That DB should be a
@@ -1543,6 +1574,8 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
   gcry_md_hd_t md = NULL;
   int algo = 0;
   size_t n;
+  
+  (void)fname;
 
   *r_crlissuer = NULL;
   *thisupdate = *nextupdate = 0;
@@ -1573,9 +1606,16 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
                 goto failure;
               }
 
-            if (opt.verbose)
+            if (opt.verbose || !*nextupdate)
               log_info (_("update times of this CRL: this=%s next=%s\n"), 
                         thisupdate, nextupdate);   
+            if (!*nextupdate)
+              {
+                log_info (_("nextUpdate not given; "
+                            "assuming a validity period of one day\n"));
+                copy_time (nextupdate, thisupdate);
+                add_isotime (nextupdate, 86400);
+              }
           }
           break;
       
@@ -1638,6 +1678,8 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
                 err = gpg_error (GPG_ERR_INV_CRL);
                 goto failure;
               }
+	    /* Note: This should be released by ksba_free, not xfree.
+	       May need a memory reallocation dance.  */
             *r_crlissuer = crlissuer; /* (Do it here so we don't need
                                          to free it later) */
 
@@ -1695,6 +1737,8 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
                            gpg_strerror (err));
                 goto failure;
               }
+	    md = NULL;
+
             err = validate_cert_chain (ctrl, crlissuer_cert,
                                        NULL, VALIDATE_MODE_CRL_RECURSIVE);
             if (err)
@@ -1719,6 +1763,8 @@ crl_parse_insert (ctrl_t ctrl, ksba_crl_t crl,
 
 
  failure:
+  if (md)
+    abort_sig_check (crl, md);
   ksba_cert_release (crlissuer_cert);
   return err;
 }
@@ -1840,12 +1886,19 @@ crl_cache_insert (ctrl_t ctrl, const char *url, ksba_reader_t reader)
   {
     char *tmpfname, *p;
     const char *nodename;
+#ifndef HAVE_W32_SYSTEM
     struct utsname utsbuf;
-    
+#endif
+
+#ifdef HAVE_W32_SYSTEM
+    nodename = "unknown";
+#else
     if (uname (&utsbuf))
       nodename = "unknown";
     else
       nodename = utsbuf.nodename;
+#endif
+
     tmpfname = xmalloc (strlen (nodename) + 100);
     sprintf (tmpfname, "crl-tmp-%s-%u-%p.db.tmp",
              nodename, (unsigned int)getpid (), tmpfname);
@@ -1879,8 +1932,9 @@ crl_cache_insert (ctrl_t ctrl, const char *url, ksba_reader_t reader)
   if (err)
     {
       log_error (_("crl_parse_insert failed: %s\n"), gpg_strerror (err));
+      /* Error in cleanup ignored.  */
+      cdb_make_finish (&cdb);
       goto leave;
-
     }
 
   /* Finish the database. */
@@ -1983,8 +2037,11 @@ crl_cache_insert (ctrl_t ctrl, const char *url, ksba_reader_t reader)
   /* Rename the temporary DB to the real name. */
   newfname = make_db_file_name (entry->issuer_hash);
   if (opt.verbose)
-    log_info (_("creating cache file `%s'\n"), newfname );
-  if (rename (fname, newfname) )
+    log_info (_("creating cache file `%s'\n"), newfname);
+#ifdef HAVE_W32_SYSTEM
+  unlink (newfname);
+#endif
+  if (rename (fname, newfname))
     {
       err = gpg_error_from_errno (errno);
       log_error (_("problem renaming `%s' to `%s': %s\n"),
