@@ -15,9 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
- * USA.
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -50,6 +48,8 @@ enum
   {
     oQuiet	  = 'q',
     oVerbose	  = 'v',
+    oLocal        = 'l',
+    oUrl          = 'u',
 
     oOCSP         = 500,
     oPing,
@@ -73,6 +73,8 @@ static ARGPARSE_OPTS opts[] = {
   { oCacheCert,"cache-cert",0, N_("add a certificate to the cache")},
   { oValidate, "validate",  0, N_("validate a certificate")},
   { oLookup,   "lookup",    0, N_("lookup a certificate")},
+  { oLocal,    "local",     0, N_("lookup only locally stored certificates")},
+  { oUrl,      "url",       0, N_("expect an URL for --lookup")},
   { oLoadCRL,  "load-crl",  0, N_("load a CRL into the dirmngr")},
   { oSquidMode,"squid-mode",0, N_("special mode for use by Squid")},
   { oPEM,      "pem",       0, N_("certificates are expected in PEM format")},
@@ -92,9 +94,10 @@ static struct
   int force_default_responder;
   int pem;
   int escaped_pem; /* PEM is additional percent encoded.  */
+  int url;         /* Expect an URL.  */
+  int local;       /* Lookup up only local certificates.  */
 
   int use_ocsp;
-  
 } opt;
 
 
@@ -231,6 +234,8 @@ main (int argc, char **argv )
         case oCacheCert: cmd_cache_cert = 1; break;
         case oValidate: cmd_validate = 1; break;
         case oLookup: cmd_lookup = 1; break;
+        case oUrl: opt.url = 1; break;
+        case oLocal: opt.local = 1; break;
         case oLoadCRL: cmd_loadcrl = 1; break;
         case oPEM: opt.pem = 1; break;
         case oSquidMode: 
@@ -370,7 +375,7 @@ main (int argc, char **argv )
         log_info (_("a dirmngr daemon is up and running\n"));
       return 0;
     }
-  else if (cmd_lookup||cmd_loadcrl||cmd_squid_mode)
+  else if (cmd_lookup|| cmd_loadcrl || cmd_squid_mode)
     return err? 1:0;
   else if (cmd_cache_cert)
     {
@@ -939,9 +944,13 @@ do_lookup (assuan_context_t ctx, const char *pattern)
   if (err)
     return err;
 
-  line = xmalloc (10 + strlen (pattern)*3 + 1);
+  line = xmalloc (10 + 6 + 13 + strlen (pattern)*3 + 1);
 
   p = stpcpy (line, "LOOKUP ");
+  if (opt.url)
+    p = stpcpy (p, "--url ");
+  if (opt.local)
+    p = stpcpy (p, "--cache-only ");
   for (s=pattern; *s; s++)
     {
       if (*s < ' ' || *s == '+')
