@@ -1,6 +1,6 @@
 /* dirmngr.c - LDAP access
  *	Copyright (C) 2002 Klarälvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004, 2006 g10 Code GmbH
+ *      Copyright (C) 2003, 2004, 2006, 2007 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -16,7 +16,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301, USA.
  */
 
 #include <config.h>
@@ -30,16 +31,18 @@
 #include <assert.h>
 #include <time.h>
 #include <fcntl.h>
+#ifndef HAVE_W32_SYSTEM
 #include <sys/socket.h>
-#include <sys/stat.h>
 #include <sys/un.h>
+#endif
+#include <sys/stat.h>
 #include <unistd.h>
 #include <signal.h>
 #include <pth.h>
 
 #include <gcrypt.h>
 #include <ksba.h>
-#include <assuan.h> /* Needed for the malloc hooks */
+#include <assuan.h> /* Needed for the malloc and socket hooks */
 
 #define JNLIB_NEED_LOG_LOGV
 #include "dirmngr.h"
@@ -48,7 +51,14 @@
 #include "crlfetch.h"
 #include "misc.h"
 #include "i18n.h"
+#include "ldapserver.h"
 
+#ifdef HAVE_W32_SYSTEM
+#define sleep _sleep
+#define FD2INT(h) ((unsigned int) (h))
+#else
+#define FD2INT(h) (h)
+#endif
 
 enum cmd_and_opt_values 
 { aNull = 0,
@@ -60,6 +70,7 @@ enum cmd_and_opt_values
   
   aServer,
   aDaemon,
+  aService,
   aListCRLs,
   aLoadCRL,
   aFetchCRL,
@@ -94,6 +105,7 @@ enum cmd_and_opt_values
   oOCSPResponder,
   oOCSPSigner,
   oOCSPMaxClockSkew,
+  oOCSPMaxPeriod,
   oOCSPCurrentPeriod,
   oMaxReplies,
   oFakedSystemTime,
@@ -112,6 +124,9 @@ static ARGPARSE_OPTS opts[] = {
 
   { aServer,   "server",    256, N_("run in server mode (foreground)") },
   { aDaemon,   "daemon",    256, N_("run in daemon mode (background)") },
+#ifdef HAVE_W32_SYSTEM
+  { aService,  "service",   256, N_("run as windows service (background)") },
+#endif
   { aListCRLs, "list-crls", 256, N_("list the contents of the CRL cache")},
   { aLoadCRL,  "load-crl",  256, N_("|FILE|load CRL from FILE into cache")},
   { aFetchCRL, "fetch-crl", 256, N_("|URL|fetch a CRL from URL")},
@@ -163,6 +178,7 @@ static ARGPARSE_OPTS opts[] = {
   { oOCSPResponder, "ocsp-responder", 2, N_("|URL|use OCSP responder at URL")},
   { oOCSPSigner, "ocsp-signer", 2, N_("|FPR|OCSP response signed by FPR")}, 
   { oOCSPMaxClockSkew, "ocsp-max-clock-skew", 1, "@" },
+  { oOCSPMaxPeriod, "ocsp-max-period", 1, "@" },
   { oOCSPCurrentPeriod, "ocsp-current-period", 1, "@" },
 
   { oMaxReplies, "max-replies", 1,
@@ -189,27 +205,45 @@ static ARGPARSE_OPTS opts[] = {
 
 #define DEFAULT_MAX_REPLIES 10
 #define DEFAULT_LDAP_TIMEOUT 100 /* arbitrary large timeout */
-#define DEFAULT_SOCKET_NAME "/var/run/dirmngr/socket"
 
 /* For the cleanup handler we need to keep track of the socket's name. */
 static const char *socket_name;
+
+/* We need to keep track of the server's nonces (these are dummies for
+   POSIX systems). */
+static assuan_sock_nonce_t socket_nonce;
+
 /* Only if this flag has been set we will remove the socket file.  */
 static int cleanup_socket;
+
 /* Keep track of the current log file so that we can avoid updating
    the log file after a SIGHUP if it didn't changed. Malloced. */
 static char *current_logfile;
+
 /* Helper to implement --debug-level. */
 static const char *debug_level;
+
 /* Flag indicating that a shutdown has been requested.  */
-static int shutdown_pending;
+static volatile int shutdown_pending;
+
 /* Counter for the active connections.  */
 static int active_connections;
 
+/* The timer tick used for housekeeping stuff.  For Windows we use a
+   longer period as the SetWaitableTimer seems to signal earlier than
+   the 2 seconds.  */
+#ifdef HAVE_W32_SYSTEM
+#define TIMERTICK_INTERVAL    (4)
+#else
+#define TIMERTICK_INTERVAL    (2)    /* Seconds.  */
+#endif
+
+
 /* Prototypes. */
 static void cleanup (void);
 static ldap_server_t parse_ldapserver_file (const char* filename);
-static void free_ldapservers_list (ldap_server_t servers);
-static void handle_connections (int listen_fd);
+static fingerprint_list_t parse_ocsp_signer (const char *string);
+static void handle_connections (assuan_fd_t listen_fd);
 
 /* Pth wrapper function definitions. */
 GCRY_THREAD_OPTION_PTH_IMPL;
@@ -226,8 +260,9 @@ my_strusage( int level )
     case 13: p = VERSION; break;
     case 14: p = "Copyright (C) " COPYRIGHT_YEAR_NAME; break;
     case 17: p = PRINTABLE_OS_NAME; break;
-      /* TRANSLATORS: @EMAIL will get replaced by the actual bug
-         reporting address. */
+      /* TRANSLATORS: @EMAIL@ will get replaced by the actual bug
+         reporting address.  This is so that we can change the
+         reporting address without breaking the translations.  */
     case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
     case 49: p = PACKAGE_BUGREPORT; break;
     case 1:
@@ -247,7 +282,9 @@ my_strusage( int level )
 static void
 my_gcry_logger (void *dummy, int level, const char *fmt, va_list arg_ptr)
 {
-  /* translate the log levels */
+  (void)dummy;
+
+  /* Translate the log levels. */
   switch (level)
     {
     case GCRY_LOG_CONT: level = JNLIB_LOG_CONT; break;
@@ -272,6 +309,8 @@ my_ksba_hash_buffer (void *arg, const char *oid,
                      const void *buffer, size_t length, size_t resultsize,
                      unsigned char *result, size_t *resultlen)
 {
+  (void)arg;
+
   if (oid && strcmp (oid, "1.3.14.3.2.26")) 
     return gpg_error (GPG_ERR_NOT_SUPPORTED); 
   if (resultsize < 20)
@@ -372,6 +411,7 @@ launch_reaper_thread (void)
                  strerror (errno) );
       dirmngr_exit (1);
     }
+  pth_attr_destroy (tattr);
 }
 
 /* Helper to stop the reaper thread for the ldap wrapper.  */
@@ -406,10 +446,16 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.ignore_ocsp_service_url = 0;
       opt.allow_ocsp = 0;
       opt.ocsp_responder = NULL;
-      opt.ocsp_signer = NULL; 
       opt.ocsp_max_clock_skew = 10 * 60;      /* 10 minutes.  */
+      opt.ocsp_max_period = 90 * 86400;       /* 90 days.  */
       opt.ocsp_current_period = 3 * 60 * 60;  /* 3 hours. */
       opt.max_replies = DEFAULT_MAX_REPLIES;
+      while (opt.ocsp_signer)
+        {
+          fingerprint_list_t tmp = opt.ocsp_signer->next;
+          xfree (opt.ocsp_signer);
+          opt.ocsp_signer = tmp;
+        }
       return 1;
     }
 
@@ -452,8 +498,11 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
 
     case oAllowOCSP: opt.allow_ocsp = 1; break;
     case oOCSPResponder: opt.ocsp_responder = pargs->r.ret_str; break;
-    case oOCSPSigner:    opt.ocsp_signer = pargs->r.ret_str; break;
+    case oOCSPSigner: 
+      opt.ocsp_signer = parse_ocsp_signer (pargs->r.ret_str);
+      break;
     case oOCSPMaxClockSkew: opt.ocsp_max_clock_skew = pargs->r.ret_int; break;
+    case oOCSPMaxPeriod: opt.ocsp_max_period = pargs->r.ret_int; break;
     case oOCSPCurrentPeriod: opt.ocsp_current_period = pargs->r.ret_int; break;
 
     case oMaxReplies: opt.max_replies = pargs->r.ret_int; break;
@@ -465,9 +514,49 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
   return 1; /* Handled. */
 }
 
-int
-main (int argc, char **argv )
+
+#ifdef HAVE_W32_SYSTEM
+/* The global status of our service.  */
+SERVICE_STATUS_HANDLE service_handle;
+SERVICE_STATUS service_status;
+
+DWORD WINAPI
+w32_service_control (DWORD control, DWORD event_type, LPVOID event_data,
+		     LPVOID context)
 {
+  /* event_type and event_data are not used here.  */
+  switch (control)
+    {
+    case SERVICE_CONTROL_SHUTDOWN:
+      /* For shutdown we will try to force termination.  */
+      service_status.dwCurrentState = SERVICE_STOP_PENDING;
+      SetServiceStatus (service_handle, &service_status); 
+      shutdown_pending = 3;
+      break;
+
+    case SERVICE_CONTROL_STOP:
+      service_status.dwCurrentState = SERVICE_STOP_PENDING;
+      SetServiceStatus (service_handle, &service_status); 
+      shutdown_pending = 1;
+      break;
+
+    default:
+      break;
+    }
+  return 0;
+}
+#endif /*HAVE_W32_SYSTEM*/
+
+
+#ifdef HAVE_W32_SYSTEM
+#define main real_main
+#endif
+int
+main (int argc, char **argv)
+{
+#ifdef HAVE_W32_SYSTEM
+#undef main
+#endif
   enum cmd_and_opt_values cmd = 0;
   ARGPARSE_ARGS pargs;
   int orig_argc;
@@ -488,6 +577,29 @@ main (int argc, char **argv )
   int rc;
   int homedir_seen = 0;
 
+#ifdef HAVE_W32_SYSTEM
+  /* The option will be set by main() below if we should run as a
+     system daemon.  */
+  if (opt.system_service)
+    {
+      service_handle
+	= RegisterServiceCtrlHandlerEx ("DirMngr",
+					&w32_service_control, NULL /*FIXME*/);
+      if (service_handle == 0)
+	log_error ("failed to register service control handler: ec=%d",
+		   (int) GetLastError ());
+      service_status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
+      service_status.dwCurrentState = SERVICE_START_PENDING;
+      service_status.dwControlsAccepted
+	= SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN;
+      service_status.dwWin32ExitCode = NO_ERROR;
+      service_status.dwServiceSpecificExitCode = NO_ERROR;
+      service_status.dwCheckPoint = 0;
+      service_status.dwWaitHint = 10000; /* 10 seconds timeout.  */
+      SetServiceStatus (service_handle, &service_status); 
+    }
+#endif /*HAVE_W32_SYSTEM*/
+
   set_strusage (my_strusage);
   log_set_prefix ("dirmngr", 1|4); 
   
@@ -539,15 +651,7 @@ main (int argc, char **argv )
   if (shell && strlen (shell) >= 3 && !strcmp (shell+strlen (shell)-3, "csh") )
     csh_style = 1;
   
-  opt.homedir = getenv("GNUPGHOME");
-  if (!opt.homedir || !*opt.homedir)
-    {
-#ifdef HAVE_DRIVE_LETTERS
-      opt.homedir = "c:/gnupg";
-#else
-      opt.homedir = "~/.gnupg";
-#endif
-    }
+  opt.homedir = default_homedir ();
 
   /* Reset rereadable options to default values. */
   parse_rereadable_options (NULL, 0); 
@@ -557,8 +661,7 @@ main (int argc, char **argv )
   opt.ldaptimeout = DEFAULT_LDAP_TIMEOUT;
 
   /* Other defaults.  */
-  socket_name = DEFAULT_SOCKET_NAME;
-
+  socket_name = default_socket_name ();
 
   /* Check whether we have a config file given on the commandline */
   orig_argc = argc;
@@ -585,6 +688,20 @@ main (int argc, char **argv )
         }
       else if (pargs.r_opt == aDaemon)
         opt.system_daemon = 1;
+      else if (pargs.r_opt == aService)
+        {
+	  /* Redundant.  The main function takes care of it.  */
+	  opt.system_service = 1;
+	  opt.system_daemon = 1;
+	}
+#ifdef HAVE_W32_SYSTEM
+      else if (pargs.r_opt == aGPGConfList || pargs.r_opt == aGPGConfTest)
+	/* We set this so we switch to the system configuration
+	   directory below.  This is a crutch to solve the problem
+	   that the user configuration is never used on Windows.  Also
+	   see below at aGPGConfList.  */
+        opt.system_daemon = 1;
+#endif
     }
 
   /* If --daemon has been given on the command line but not --homedir,
@@ -592,9 +709,9 @@ main (int argc, char **argv )
      this also overrides the GNUPGHOME environment variable.  */
   if (opt.system_daemon && !homedir_seen)
     {
-      opt.homedir = DIRMNGR_SYSCONFDIR;
-      opt.homedir_data = DIRMNGR_DATADIR;
-      opt.homedir_cache = DIRMNGR_CACHEDIR;
+      opt.homedir = dirmngr_sysconfdir ();
+      opt.homedir_data = dirmngr_datadir ();
+      opt.homedir_cache = dirmngr_cachedir ();
     }
 
   if (default_config)
@@ -640,6 +757,7 @@ main (int argc, char **argv )
         {
         case aServer: 
         case aDaemon:
+        case aService:
         case aShutdown: 
         case aFlush: 
 	case aListCRLs: 
@@ -747,15 +865,18 @@ main (int argc, char **argv )
   else
       opt.ldapservers = parse_ldapserver_file (ldapfile);
 
+#ifndef HAVE_W32_SYSTEM
   /* We need to ignore the PIPE signal because the we might log to a
      socket and that code handles EPIPE properly.  The ldap wrapper
      also requires us to ignore this silly signal. Assuan would set
      this signal to ignore anyway.*/
   signal (SIGPIPE, SIG_IGN);
-
+#endif
 
   /* Ready.  Now to our duties. */
-  if (!cmd)
+  if (!cmd && opt.system_service)
+    cmd = aDaemon;
+  else if (!cmd)
     cmd = aServer;
   rc = 0;
 
@@ -768,6 +889,7 @@ main (int argc, char **argv )
         {
           log_set_file (logfile);
           log_set_prefix (NULL, 2|4);
+	  assuan_set_assuan_log_stream (log_get_stream ());
         }
        
       if (debug_wait)
@@ -781,12 +903,12 @@ main (int argc, char **argv )
       launch_reaper_thread ();
       cert_cache_init ();
       crl_cache_init ();
-      start_command_handler (-1);
+      start_command_handler (ASSUAN_INVALID_FD);
       shutdown_reaper ();
     }
   else if (cmd == aDaemon)
     {
-      int fd;
+      assuan_fd_t fd;
       pid_t pid;
       int len;
       struct sockaddr_un serv_addr;
@@ -802,23 +924,26 @@ main (int argc, char **argv )
                                  |JNLIB_LOG_WITH_TIME
                                  |JNLIB_LOG_WITH_PID));
           current_logfile = xstrdup (logfile);
+	  assuan_set_assuan_log_stream (log_get_stream ());
         }
 
-      if (strchr (socket_name, ':') )
+#ifndef HAVE_W32_SYSTEM
+      if (strchr (socket_name, ':'))
         {
           log_error (_("colons are not allowed in the socket name\n"));
           dirmngr_exit (1);
         }
+#endif
       if (strlen (socket_name)+1 >= sizeof serv_addr.sun_path ) 
         {
           log_error (_("name of socket too long\n"));
           dirmngr_exit (1);
         }
-   
-      fd = socket (AF_UNIX, SOCK_STREAM, 0);
-      if (fd == -1)
+    
+      fd = assuan_sock_new (AF_UNIX, SOCK_STREAM, 0);
+      if (fd == ASSUAN_INVALID_FD)
         {
-          log_error (_("can't create socket: %s\n"), strerror(errno) );
+          log_error (_("can't create socket: %s\n"), strerror (errno));
           cleanup ();
           dirmngr_exit (1);
         }
@@ -827,27 +952,30 @@ main (int argc, char **argv )
       serv_addr.sun_family = AF_UNIX;
       strcpy (serv_addr.sun_path, socket_name);
       len = (offsetof (struct sockaddr_un, sun_path)
-             + strlen(serv_addr.sun_path) + 1);
+             + strlen (serv_addr.sun_path) + 1);
 
-      rc = bind (fd, (struct sockaddr*)&serv_addr, len);
+      rc = assuan_sock_bind (fd, (struct sockaddr*) &serv_addr, len);
       if (rc == -1 && errno == EADDRINUSE)
-        {
-          remove (socket_name);
-          rc = bind (fd, (struct sockaddr*)&serv_addr, len);
-        }
+	{
+	  remove (socket_name);
+	  rc = assuan_sock_bind (fd, (struct sockaddr*) &serv_addr, len);
+	}
+      if (rc != -1 
+	  && (rc = assuan_sock_get_nonce ((struct sockaddr*) &serv_addr, len, &socket_nonce)))
+	log_error (_("error getting nonce for the socket\n"));
       if (rc == -1)
         {
           log_error (_("error binding socket to `%s': %s\n"),
-                     serv_addr.sun_path, strerror (errno) );
-          close (fd);
+                     serv_addr.sun_path, gpg_strerror (gpg_error_from_errno (errno)));
+          assuan_sock_close (fd);
           dirmngr_exit (1);
         }
       cleanup_socket = 1;
   
-      if (listen (fd, 5 ) == -1)
+      if (listen (FD2INT (fd), 5) == -1)
         {
           log_error (_("listen() failed: %s\n"), strerror (errno));
-          close (fd);
+          assuan_sock_close (fd);
           dirmngr_exit (1);
         }
 
@@ -855,6 +983,11 @@ main (int argc, char **argv )
         log_info (_("listening on socket `%s'\n"), socket_name );
 
       fflush (NULL);
+
+#ifdef HAVE_W32_SYSTEM
+      pid = getpid ();
+      printf ("set DIRMNGR_INFO=%s;%lu;1\n", socket_name, (ulong) pid);
+#else
       pid = pth_fork ();
       if (pid == (pid_t)-1) 
         {
@@ -929,13 +1062,28 @@ main (int argc, char **argv )
               dirmngr_exit (1);
             }
         }
+#endif
 
       launch_reaper_thread ();
       cert_cache_init ();
       crl_cache_init ();
+#ifdef HAVE_W32_SYSTEM
+      if (opt.system_service)
+	{
+	  service_status.dwCurrentState = SERVICE_RUNNING;
+	  SetServiceStatus (service_handle, &service_status);
+	}
+#endif
       handle_connections (fd);
-      close (fd);
+      assuan_sock_close (fd);
       shutdown_reaper ();
+#ifdef HAVE_W32_SYSTEM
+      if (opt.system_service)
+	{
+	  service_status.dwCurrentState = SERVICE_STOPPED;
+	  SetServiceStatus (service_handle, &service_status);
+	}
+#endif
     }
   else if (cmd == aListCRLs)
     {
@@ -1002,7 +1150,10 @@ main (int argc, char **argv )
     dirmngr_exit (0);
   else if (cmd == aGPGConfList)
     {
+      unsigned long flags = 0;
       char *filename;
+      char *filename_esc;
+
       /* List options and default values in the GPG Conf format.  */
 
 /* The following list is taken from gnupg/tools/gpgconf-comp.c.  */
@@ -1018,6 +1169,16 @@ main (int argc, char **argv )
 /* The NO_ARG_DESC flag for an option indicates that the argument has
    a default, which is described by the value of the ARGDEF field.  */
 #define GC_OPT_FLAG_NO_ARG_DESC	(1UL << 6)
+#define GC_OPT_FLAG_NO_CHANGE   (1UL <<7)
+
+#ifdef HAVE_W32_SYSTEM
+      /* On Windows systems, dirmngr always runs as system daemon, and
+	 the per-user configuration is never used.  So we short-cut
+	 everything to use the global system configuration of dirmngr
+	 above, and here we set the no change flag to make these
+	 read-only.  */
+      flags |= GC_OPT_FLAG_NO_CHANGE;
+#endif
 
       /* First the configuration file.  This is not an option, but it
 	 is vital information for GPG Conf.  */
@@ -1025,14 +1186,16 @@ main (int argc, char **argv )
         opt.config_filename = make_filename (opt.homedir,
                                              "dirmngr.conf", NULL );
 
+      filename = percent_escape (opt.config_filename, NULL);
       printf ("gpgconf-dirmngr.conf:%lu:\"%s\n",
-              GC_OPT_FLAG_DEFAULT, opt.config_filename);
+              GC_OPT_FLAG_DEFAULT, filename);
+      xfree (filename);
 
-      printf ("verbose:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("quiet:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("debug-level:%lu:\"none\n", GC_OPT_FLAG_DEFAULT);
-      printf ("log-file:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("force:%lu:\n", GC_OPT_FLAG_NONE);
+      printf ("verbose:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("quiet:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("debug-level:%lu:\"none\n", flags | GC_OPT_FLAG_DEFAULT);
+      printf ("log-file:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("force:%lu:\n", flags | GC_OPT_FLAG_NONE);
 
       /* --csh and --sh are mutually exclusive, something we can not
          express in GPG Conf.  --options is only usable from the
@@ -1044,45 +1207,82 @@ main (int argc, char **argv )
                                 opt.system_daemon?
                                 "ldapservers.conf":"dirmngr_ldapservers.conf",
                                 NULL);
-      printf ("ldapserverlist-file:%lu:\"%s\n", GC_OPT_FLAG_DEFAULT, filename);
+      filename_esc = percent_escape (filename, NULL);
+      printf ("ldapserverlist-file:%lu:\"%s\n", GC_OPT_FLAG_DEFAULT,
+	      filename_esc);
+      xfree (filename_esc);
       xfree (filename);
 
       printf ("ldaptimeout:%lu:%u\n",
-              GC_OPT_FLAG_DEFAULT, DEFAULT_LDAP_TIMEOUT);
+              flags | GC_OPT_FLAG_DEFAULT, DEFAULT_LDAP_TIMEOUT);
       printf ("max-replies:%lu:%u\n",
-              GC_OPT_FLAG_DEFAULT, DEFAULT_MAX_REPLIES);
-      printf ("allow-ocsp:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("ocsp-responder:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("ocsp-signer:%lu:\n", GC_OPT_FLAG_NONE);
+              flags | GC_OPT_FLAG_DEFAULT, DEFAULT_MAX_REPLIES);
+      printf ("allow-ocsp:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("ocsp-responder:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("ocsp-signer:%lu:\n", flags | GC_OPT_FLAG_NONE);
 
-      printf ("faked-system-time:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("no-greeting:%lu:\n", GC_OPT_FLAG_NONE);
+      printf ("faked-system-time:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("no-greeting:%lu:\n", flags | GC_OPT_FLAG_NONE);
 
-      printf ("disable-http:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("disable-ldap:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("honor-http-proxy:%lu\n", GC_OPT_FLAG_NONE);
-      printf ("http-proxy:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("ldap-proxy:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("only-ldap-proxy:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("ignore-ldap-dp:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("ignore-http-dp:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("ignore-ocsp-service-url:%lu:\n", GC_OPT_FLAG_NONE);
+      printf ("disable-http:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("disable-ldap:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("honor-http-proxy:%lu\n", flags | GC_OPT_FLAG_NONE);
+      printf ("http-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("ldap-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("only-ldap-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("ignore-ldap-dp:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("ignore-http-dp:%lu:\n", flags | GC_OPT_FLAG_NONE);
+      printf ("ignore-ocsp-service-url:%lu:\n", flags | GC_OPT_FLAG_NONE);
       /* Note: The next one is to fix a typo in gpgconf - should be
          removed eventually. */
-      printf ("ignore-ocsp-servic-url:%lu:\n", GC_OPT_FLAG_NONE);
+      printf ("ignore-ocsp-servic-url:%lu:\n", flags | GC_OPT_FLAG_NONE);
     }
   cleanup ();
   return !!rc;
 }
 
 
+#ifdef HAVE_W32_SYSTEM
+int
+main (int argc, char *argv[])
+{
+  int i;
+
+  /* Find out if we run in daemon mode or on the command line.  */
+  for (i = 1; i < argc; i++)
+    if (!strcmp (argv[i], "--service"))
+      {
+	opt.system_service = 1;
+	opt.system_daemon = 1;
+	break;
+      }
+
+  if (!opt.system_service)
+    return real_main (argc, argv);
+  else
+    {
+      SERVICE_TABLE_ENTRY DispatchTable [] =
+	{
+	  /* Ignore warning.  */
+	  { "DirMngr", &real_main },
+	  { NULL, NULL }
+	};
+
+      if (!StartServiceCtrlDispatcher (DispatchTable))
+        return 1;
+      return 0;
+    }
+}
+#endif
+
+
 static void
 cleanup (void)
 {
   crl_cache_deinit ();
   cert_cache_deinit (1);
 
-  free_ldapservers_list (opt.ldapservers);
+  ldapserver_list_free (opt.ldapservers);
   opt.ldapservers = NULL;
 
   if (cleanup_socket)
@@ -1105,6 +1305,8 @@ dirmngr_exit (int rc)
 void
 dirmngr_init_default_ctrl (ctrl_t ctrl)
 {
+  (void)ctrl;
+
   /* Nothing for now. */
 }
 
@@ -1127,12 +1329,11 @@ static ldap_server_t
 parse_ldapserver_file (const char* filename)
 {
   char buffer[1024];
-  char *p, *endp;
+  char *p;
   ldap_server_t server, serverstart, *serverend;
   int c;
   unsigned int lineno = 0;
   FILE *fp;
-  int fieldno;
 
   fp = fopen (filename, "r");
   if (!fp)
@@ -1145,8 +1346,6 @@ parse_ldapserver_file (const char* filename)
   serverend = &serverstart;
   while (fgets (buffer, sizeof buffer, fp))
     {
-      int fail = 0;
-
       lineno++;
       if (!*buffer || buffer[strlen(buffer)-1] != '\n')
         {
@@ -1168,64 +1367,8 @@ parse_ldapserver_file (const char* filename)
         continue;
 
       /* Parse the colon separated fields. */
-      server = xcalloc (1, sizeof *server);
-      for (fieldno=1, p = buffer; p; p = endp, fieldno++ )
-        {
-          endp = strchr (p, ':');
-          if (endp)
-            *endp++ = '\0';
-          trim_spaces (p);
-          switch (fieldno)
-            {
-            case 1:
-              if (*p)
-                server->host = xstrdup (p);
-              else
-                {
-                  log_error (_("%s:%u: no hostname given\n"),
-                             filename, lineno);
-                  fail = 1;
-                }
-              break;
-                
-            case 2:
-              if (*p)
-                server->port = atoi (p);
-              break;
-
-            case 3:
-              if (*p)
-                server->user = xstrdup (p);
-              break;
-
-            case 4:
-              if (*p && !server->user)
-                {
-                  log_error (_("%s:%u: password given without user\n"), 
-                             filename, lineno);
-                  fail = 1;
-                }
-              else if (*p)
-                server->pass = xstrdup (p);
-              break;
-
-            case 5:
-              if (*p)
-                server->base = xstrdup (p);
-              break;
-
-            default:
-              /* (We silently ignore extra fields.) */
-              break;
-            }
-        }
-          
-      if (fail)
-        {
-          log_info (_("%s:%u: skipping this line\n"), filename, lineno);
-          free_ldapservers_list (server);
-        }
-      else
+      server = ldapserver_parse_one (buffer, filename, lineno);
+      if (server)
         {
           *serverend = server;
           serverend = &server->next;
@@ -1240,26 +1383,129 @@ parse_ldapserver_file (const char* filename)
 }
 
 
-/* Release the list of SERVERS. As usual it is okay to call this
-   fucntion with SERVERS passed as NULL. */
-static void 
-free_ldapservers_list (ldap_server_t servers)
+static fingerprint_list_t
+parse_ocsp_signer (const char *string)
 {
-  while (servers)
+  gpg_error_t err;
+  char *fname;
+  FILE *fp;
+  char line[256];
+  char *p;
+  fingerprint_list_t list, *list_tail, item;
+  unsigned int lnr = 0;
+  int c, i, j;
+  int errflag = 0;
+
+
+  /* Check whether this is not a filename and treat it as a direct
+     fingerprint specification.  */
+  if (!strpbrk (string, "/.~\\"))
     {
-      ldap_server_t tmp = servers->next;
-      xfree (servers->host);
-      xfree (servers->user);
-      if (servers->pass)
-        memset (servers->pass, 0, strlen (servers->pass));
-      xfree (servers->pass);
-      xfree (servers->base);
-      xfree (servers);
-      servers = tmp;
+      item = xcalloc (1, sizeof *item);
+      for (i=j=0; (string[i] == ':' || hexdigitp (string+i)) && j < 40; i++)
+        if ( string[i] != ':' )
+          item->hexfpr[j++] = string[i] >= 'a'? (string[i] & 0xdf): string[i];
+      item->hexfpr[j] = 0;
+      if (j != 40 || !(spacep (string+i) || !string[i]))
+        {
+          log_error (_("%s:%u: invalid fingerprint detected\n"), 
+                     "--ocsp-signer", 0);
+          xfree (item);
+          return NULL;
+        }
+      return item;
     }
+  
+  /* Well, it is a filename.  */
+  if (*string == '/' || (*string == '~' && string[1] == '/'))
+    fname = make_filename (string, NULL);
+  else
+    {
+      if (string[0] == '.' && string[1] == '/' )
+        string += 2;
+      fname = make_filename (opt.homedir, string, NULL);
+    }
+
+  fp = fopen (fname, "r");
+  if (!fp)
+    {
+      err = gpg_error_from_syserror ();
+      log_error (_("can't open `%s': %s\n"), fname, gpg_strerror (err));
+      xfree (fname);
+      return NULL;
+    }
+
+  list = NULL;
+  list_tail = &list;
+  for (;;)
+    {
+      if (!fgets (line, DIM(line)-1, fp) )
+        {
+          if (!feof (fp))
+            {
+              err = gpg_error_from_syserror ();
+              log_error (_("%s:%u: read error: %s\n"),
+                         fname, lnr, gpg_strerror (err));
+              errflag = 1;
+            }
+          fclose (fp);
+          if (errflag)
+            {
+              while (list)
+                {
+                  fingerprint_list_t tmp = list->next;
+                  xfree (list);
+                  list = tmp;
+                }
+            }
+          xfree (fname);
+          return list; /* Ready.  */
+        }
+
+      lnr++;
+      if (!*line || line[strlen(line)-1] != '\n')
+        {
+          /* Eat until end of line. */
+          while ( (c=getc (fp)) != EOF && c != '\n')
+            ;
+          err = gpg_error (*line? GPG_ERR_LINE_TOO_LONG
+                           /* */: GPG_ERR_INCOMPLETE_LINE);
+          log_error (_("%s:%u: read error: %s\n"), 
+                     fname, lnr, gpg_strerror (err));
+          errflag = 1;
+          continue;
+        }
+
+      /* Allow for empty lines and spaces */
+      for (p=line; spacep (p); p++)
+        ;
+      if (!*p || *p == '\n' || *p == '#')
+        continue;
+
+      item = xcalloc (1, sizeof *item);
+      *list_tail = item;
+      list_tail = &item->next;
+
+      for (i=j=0; (p[i] == ':' || hexdigitp (p+i)) && j < 40; i++)
+        if ( p[i] != ':' )
+          item->hexfpr[j++] = p[i] >= 'a'? (p[i] & 0xdf): p[i];
+      item->hexfpr[j] = 0;
+      if (j != 40 || !(spacep (p+i) || p[i] == '\n'))
+        {
+          log_error (_("%s:%u: invalid fingerprint detected\n"), fname, lnr);
+          errflag = 1;
+        }
+      i++;
+      while (spacep (p+i))
+        i++;
+      if (p[i] && p[i] != '\n')
+        log_info (_("%s:%u: garbage at end of line ignored\n"), fname, lnr);
+    }
+  /*NOTREACHED*/
 }
 
 
+
 
 /*
    Stuff used in daemon mode.  
@@ -1321,6 +1567,7 @@ handle_signal (int signo)
 {
   switch (signo)
     {
+#ifndef HAVE_W32_SYSTEM
     case SIGHUP:
       log_info (_("SIGHUP received - "
                   "re-reading configuration and flushing caches\n"));
@@ -1361,28 +1608,69 @@ handle_signal (int signo)
       cleanup ();
       dirmngr_exit (0);
       break;
-
+#endif
     default:
       log_info (_("signal %d received - no action defined\n"), signo);
     }
 }
 
 
+/* This is the worker for the ticker.  It is called every few seconds
+   and may only do fast operations. */
+static void
+handle_tick (void)
+{
+  /* Nothing real to do right now.  Actually we need the timeout only
+     for W32 where we don't use signals and need a way for the loop to
+     check for the shutdown flag. */
+#ifdef HAVE_W32_SYSTEM
+  if (shutdown_pending)
+    log_info (_("SIGTERM received - shutting down ...\n"));
+  if (shutdown_pending > 2)
+    {
+      log_info (_("shutdown forced\n"));
+      log_info ("%s %s stopped\n", strusage(11), strusage(13) );
+      cleanup ();
+      dirmngr_exit (0);
+    }
+#endif /*HAVE_W32_SYSTEM*/
+}
+
+
+/* Check the nonce on a new connection.  This is a NOP unless we we
+   are using our Unix domain socket emulation under Windows.  */
+static int 
+check_nonce (assuan_fd_t fd, assuan_sock_nonce_t *nonce)
+{
+  if (assuan_sock_check_nonce (fd, nonce))
+    {
+      log_info (_("error reading nonce on fd %d: %s\n"), 
+                FD2INT (fd), strerror (errno));
+      assuan_sock_close (fd);
+      return -1;
+    }
+  else
+    return 0;
+}
+
 
 /* Helper to call a connection's main fucntion. */
 static void *
 start_connection_thread (void *arg)
 {
-  int fd = (int)arg;
+  assuan_fd_t fd = (assuan_fd_t) arg;
+
+  if (check_nonce (fd, &socket_nonce))
+    return NULL;
 
   active_connections++;
   if (opt.verbose)
-    log_info (_("handler for fd %d started\n"), fd);
+    log_info (_("handler for fd %d started\n"), FD2INT (fd));
 
   start_command_handler (fd);
 
   if (opt.verbose)
-    log_info (_("handler for fd %d terminated\n"), fd);
+    log_info (_("handler for fd %d terminated\n"), FD2INT (fd));
   active_connections--;
   
   return NULL;
@@ -1391,21 +1679,22 @@ start_connection_thread (void *arg)
 
 /* Main loop in daemon mode. */
 static void
-handle_connections (int listen_fd)
+handle_connections (assuan_fd_t listen_fd)
 {
   pth_attr_t tattr;
-  pth_event_t ev;
-  sigset_t sigs;
+  pth_event_t ev, time_ev;
+  sigset_t sigs, oldsigs;
   int signo;
   struct sockaddr_un paddr;
   socklen_t plen = sizeof( paddr );
-  int fd;
+  assuan_fd_t fd;
 
   tattr = pth_attr_new();
   pth_attr_set (tattr, PTH_ATTR_JOINABLE, 0);
   pth_attr_set (tattr, PTH_ATTR_STACK_SIZE, 1024*1024);
   pth_attr_set (tattr, PTH_ATTR_NAME, "dirmngr");
 
+#ifndef HAVE_W32_SYSTEM /* FIXME */
   sigemptyset (&sigs );
   sigaddset (&sigs, SIGHUP);
   sigaddset (&sigs, SIGUSR1);
@@ -1413,6 +1702,11 @@ handle_connections (int listen_fd)
   sigaddset (&sigs, SIGINT);
   sigaddset (&sigs, SIGTERM);
   ev = pth_event (PTH_EVENT_SIGS, &sigs, &signo);
+#else
+  sigs = 0;
+  ev = pth_event (PTH_EVENT_SIGS, &sigs, &signo);
+#endif
+  time_ev = NULL;
 
   for (;;)
     {
@@ -1430,12 +1724,29 @@ handle_connections (int listen_fd)
           continue;
 	}
 
-      fd = pth_accept_ev (listen_fd, (struct sockaddr *)&paddr, &plen, ev);
-      if (fd == -1)
+      if (!time_ev)
+        time_ev = pth_event (PTH_EVENT_TIME, 
+                             pth_timeout (TIMERTICK_INTERVAL, 0));
+
+      if (time_ev)
+        pth_event_concat (ev, time_ev, NULL);
+      fd = (assuan_fd_t) pth_accept_ev (FD2INT (listen_fd), (struct sockaddr *)&paddr, &plen, ev);
+      if (time_ev)
+        pth_event_isolate (time_ev);
+
+      if (fd == ASSUAN_INVALID_FD)
         {
-          if (pth_event_occurred (ev))
+          if (pth_event_occurred (ev)
+              || (time_ev && pth_event_occurred (time_ev)) )
             {
-              handle_signal (signo);
+              if (pth_event_occurred (ev))
+                handle_signal (signo);
+              if (time_ev && pth_event_occurred (time_ev))
+                {
+                  pth_event_free (time_ev, PTH_FREE_ALL);
+                  time_ev = NULL;
+                  handle_tick ();
+                }
               continue;
 	    }
           log_error (_("accept failed: %s - waiting 1s\n"), strerror (errno));
@@ -1443,15 +1754,40 @@ handle_connections (int listen_fd)
           continue;
 	}
 
+      if (pth_event_occurred (ev))
+        {
+          handle_signal (signo);
+        }
+
+      if (time_ev && pth_event_occurred (time_ev))
+        {
+          pth_event_free (time_ev, PTH_FREE_ALL);
+          time_ev = NULL;
+          handle_tick ();
+        }
+
+
+      /* We now might create a new thread and because we don't want
+         any signals (as we are handling them here) to be delivered to
+         a new thread we need to block those signals. */
+      pth_sigmask (SIG_BLOCK, &sigs, &oldsigs);
+
+      /* Create thread to handle this connection.  */
       if (!pth_spawn (tattr, start_connection_thread, (void*)fd))
         {
           log_error (_("error spawning connection handler: %s\n"),
                      strerror (errno) );
-          close (fd);
+          assuan_sock_close (fd);
 	}
+
+      /* Restore the signal mask. */
+      pth_sigmask (SIG_SETMASK, &oldsigs, NULL);
     }
   
   pth_event_free (ev, PTH_FREE_ALL);
+  if (time_ev)
+    pth_event_free (time_ev, PTH_FREE_ALL);
+  pth_attr_destroy (tattr);
   cleanup ();
   log_info ("%s %s stopped\n", strusage(11), strusage(13));
 }
