@@ -15,8 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #ifndef DIRMNGR_H
@@ -27,6 +26,7 @@
 #include <gcrypt.h> /* we need this for the memory function protos */
 #include <gpg-error.h>
 #include <ksba.h>
+#include <assuan.h>
 
 /* to pass hash functions to libksba we need to cast it */
 #define HASH_FNC ((void (*)(void *, const void*,size_t))gcry_md_write)
@@ -51,6 +51,15 @@ struct ldap_server_s {
 typedef struct ldap_server_s *ldap_server_t;
 
 
+/* A list of fingerprints.  */
+struct fingerprint_list_s;
+typedef struct fingerprint_list_s *fingerprint_list_t;
+struct fingerprint_list_s
+{
+  fingerprint_list_t next;
+  char hexfpr[20+20+1];
+};
+
 
 /* A large struct name "opt" to keep global flags. */
 struct 
@@ -72,6 +81,7 @@ struct
   char *http_wrapper_program; /* Override value for the HTTP wrapper
                                  program.  */
 
+  int system_service;   /* We are running as W32 service (implies daemon).  */
   int system_daemon;    /* We are running in system daemon mode.  */
   int running_detached; /* We are running in detached mode.  */
 
@@ -96,12 +106,15 @@ struct
   ldap_server_t ldapservers;
   int add_new_ldapservers;
 
-  const char *ocsp_responder; /* Standard OCSP responder's URL. */
-  const char *ocsp_signer;    /* The fingerprint of the standard OCSP
-                                 responder signer's certificate. */
-  unsigned int ocsp_max_clock_skew;  /* Allowed seconds of clocks skew. */
-  unsigned int ocsp_current_period;  /* Seconds a response is
-                                        considered current. */
+  const char *ocsp_responder;     /* Standard OCSP responder's URL. */
+  fingerprint_list_t ocsp_signer; /* The list of fingerprints with allowed
+                                     standard OCSP signer certificates.  */
+  
+  unsigned int ocsp_max_clock_skew; /* Allowed seconds of clocks skew. */
+  unsigned int ocsp_max_period;     /* Seconds a response is at maximum
+                                       considered valid after thisUpdate. */
+  unsigned int ocsp_current_period; /* Seconds a response is considered 
+                                       current after nextUpdate. */
 } opt;
 
 
@@ -132,7 +145,15 @@ typedef struct cert_ref_s *cert_ref_t;
 
 
 /* Control structure per connection. */
-struct server_local_s;
+struct server_local_s 
+{
+  /* Data used to associate an Assuan context with local server data */
+  assuan_context_t assuan_ctx;
+
+  /* Per-session LDAP serfver.  */
+  ldap_server_t ldapservers;
+};
+
 
 struct server_control_s
 {
@@ -145,6 +166,8 @@ struct server_control_s
   int check_revocations_nest_level; /* Internal to check_revovations.  */
   cert_ref_t ocsp_certs; /* Certificates from the current OCSP
                             response. */
+
+  int audit_events;  /* Send audit events to client.  */
 };
 
 typedef struct server_control_s *ctrl_t;
@@ -159,7 +182,7 @@ ksba_cert_t get_cert_local (ctrl_t ctrl, const char *issuer);
 ksba_cert_t get_issuing_cert_local (ctrl_t ctrl, const char *issuer);
 ksba_cert_t get_cert_local_ski (ctrl_t ctrl,
                                 const char *name, ksba_sexp_t keyid);
-void start_command_handler (int fd);
+void start_command_handler (assuan_fd_t fd);
 gpg_error_t dirmngr_status (ctrl_t ctrl, const char *keyword, ...);
 gpg_error_t dirmngr_tick (ctrl_t ctrl);
 
