@@ -1,5 +1,5 @@
 /* estream.c - Extended Stream I/O Library
- * Copyright (C) 2004, 2005, 2006, 2007 g10 Code GmbH
+ * Copyright (C) 2004, 2005, 2006, 2007, 2008 g10 Code GmbH
  *
  * This file is part of Libestream.
  *
@@ -111,8 +111,12 @@ typedef pth_mutex_t estream_mutex_t;
   pth_mutex_release (&(mutex))
 # define ESTREAM_MUTEX_TRYLOCK(mutex)     \
   ((pth_mutex_acquire (&(mutex), 1, NULL) == TRUE) ? 0 : -1)
-# define ESTREAM_MUTEX_INITIALIZE(mutex)  \
-  pth_mutex_init    (&(mutex))
+# define ESTREAM_MUTEX_INITIALIZE(mutex) pth_mutex_init (&(mutex))
+#ifdef HAVE_W32_SYSTEM
+# define ESTREAM_MUTEX_DESTROY(mutex) pth_mutex_destroy (&(mutex))
+#else
+# define ESTREAM_MUTEX_DESTROY(mutex) (void) 0
+#endif
 #else
 
 typedef void *estream_mutex_t;
@@ -121,6 +125,7 @@ typedef void *estream_mutex_t;
 # define ESTREAM_MUTEX_UNLOCK(mutex) (void) 0
 # define ESTREAM_MUTEX_TRYLOCK(mutex) 0
 # define ESTREAM_MUTEX_INITIALIZE(mutex) (void) 0
+# define ESTREAM_MUTEX_DESTROY(mutex) (void) 0
 #endif
 
 /* Memory allocator functions.  */
@@ -1180,7 +1185,10 @@ es_create (estream_t *stream, void *cookie, int fd,
 
   err = es_list_add (stream_new);
   if (err)
-    goto out;
+    {
+      ESTREAM_MUTEX_DESTROY (stream_new->intern->lock);
+      goto out;
+    }
 
   *stream = stream_new;
 
@@ -1208,6 +1216,7 @@ es_destroy (estream_t stream)
     {
       es_list_remove (stream);
       err = es_deinitialize (stream);
+      ESTREAM_MUTEX_DESTROY (stream->intern->lock);
       ES_MEM_FREE (stream->intern);
       ES_MEM_FREE (stream);
     }
