@@ -1,6 +1,6 @@
 /* ldap.c - LDAP access
- *      Copyright (C) 2002 Klarälvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004, 2005 g10 Code GmbH
+ * Copyright (C) 2002 Klarälvdalens Datakonsult AB
+ * Copyright (C) 2003, 2004, 2005, 2007, 2008 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -27,16 +27,23 @@
 #include <errno.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <sys/wait.h>
 #include <time.h>
-#include <ldap.h> /* fixme: remove it */
 #include <pth.h>
 #include <ksba.h> 
 
+#include "exechelp.h"
 #include "crlfetch.h"
 #include "dirmngr.h"
+#include "ldapserver.h"
 #include "misc.h"
 
+#ifdef HAVE_W32_SYSTEM
+#define setenv(a,b,c) SetEnvironmentVariable ((a),(b))
+#else
+#define pth_close(fd) close(fd)
+#endif
+
+
 /* In case sysconf does not return a value we need to have a limit. */
 #ifdef _POSIX_OPEN_MAX
 #define MAX_OPEN_FDS _POSIX_OPEN_MAX
@@ -52,6 +59,7 @@
                             "$-_.+!*'(),"
 #define USERCERTIFICATE "userCertificate"
 #define CACERTIFICATE   "caCertificate"
+#define X509CACERT      "x509caCert"
 #define USERSMIMECERTIFICATE "userSMIMECertificate"
 
 
@@ -77,6 +85,7 @@ struct wrapper_context_s
   gpg_error_t fd_error; /* Set to the gpg_error of the last read error
                            if any.  */
   int log_fd;   /* Connected with stderr of the ldap wrapper.  */
+  pth_event_t log_ev;
   ctrl_t ctrl;  /* Connection data. */
   int ready;    /* Internally used to mark to be removed contexts. */
   ksba_reader_t reader; /* The ksba reader object or NULL. */
@@ -159,12 +168,17 @@ static void
 destroy_wrapper (struct wrapper_context_s *ctx) 
 {
   if (ctx->pid != (pid_t)(-1))
-    kill (ctx->pid, SIGTERM);
+    {
+      dirmngr_kill_process (ctx->pid);
+      dirmngr_release_process (ctx->pid);
+    }
   ksba_reader_release (ctx->reader);
-  if (ctx->fd)
-    close (ctx->fd);
-  if (ctx->log_fd)
-    close (ctx->log_fd);
+  if (ctx->fd != -1)
+    pth_close (ctx->fd);
+  if (ctx->log_fd != -1)
+    pth_close (ctx->log_fd);
+  if (ctx->log_ev)
+    pth_event_free (ctx->log_ev, PTH_FREE_THIS);
   xfree (ctx->line);
   xfree (ctx);
 }
@@ -235,55 +249,69 @@ print_log_line (struct wrapper_context_s *ctx, char *line)
 void *
 ldap_wrapper_thread (void *dummy)
 {
-  fd_set read_fds;
-  int nfds, n;
-  struct timeval tv;
-  struct wrapper_context_s *ctx, *ctx_prev;
+  int nfds;
+  int n;
+  struct wrapper_context_s *ctx;
+  struct wrapper_context_s *ctx_prev;
   char line[256];
   time_t current_time;
 
+  (void)dummy;
+
+
   for (;;)
     {
-      FD_ZERO (&read_fds);
-      
-      for (nfds = -1, ctx = wrapper_list; ctx; ctx = ctx->next)
+      pth_event_t timeout_ev;
+
+      timeout_ev = pth_event (PTH_EVENT_TIME, pth_timeout (1, 0));
+      if (! timeout_ev)
+	{
+          log_error (_("pth_event failed: %s\n"), strerror (errno));
+          pth_sleep (10);
+	  continue;
+	}
+
+      for (ctx = wrapper_list; ctx; ctx = ctx->next)
         {
           if (ctx->log_fd != -1)
             {
-              FD_SET (ctx->log_fd, &read_fds);
-              if (ctx->log_fd > nfds)
-                nfds = ctx->log_fd;
+	      pth_event_isolate (ctx->log_ev);
+	      pth_event_concat (timeout_ev, ctx->log_ev, NULL);
             }
         }
-      nfds++;
 
-      tv.tv_sec = 1;
-      tv.tv_usec = 0;
-      nfds = pth_select (nfds, &read_fds, NULL, NULL, &tv);
-      if ( nfds < 0 )
+      /* Note that the read FDs are actually handles.  Thus, we can
+	 not use pth_select, but have to use pth_wait.  */
+      nfds = pth_wait (timeout_ev);
+      if (nfds < 0)
         {
-          log_error (_("select failed: %s\n"), strerror (errno));
+          pth_event_free (timeout_ev, PTH_FREE_THIS);
+          log_error (_("pth_wait failed: %s\n"), strerror (errno));
           pth_sleep (10);
-          continue;
+	  continue;
         }
+      if (pth_event_status (timeout_ev) == PTH_STATUS_OCCURRED)
+	nfds--;
+      pth_event_free (timeout_ev, PTH_FREE_THIS);
 
       current_time = time (NULL);
       if (current_time > INACTIVITY_TIMEOUT)
         current_time -= INACTIVITY_TIMEOUT;
 
       /* Note that there is no need to lock the list because we always
-         add entries at the head and thus traversing the list will
-         even work if we have a context switch in waitpid (which should
-         anyway only happen with Pth's hard system call mapping).   */
+         add entries at the head (with a pending event status) and
+         thus traversing the list will even work if we have a context
+         switch in waitpid (which should anyway only happen with Pth's
+         hard system call mapping).  */
       for (ctx = wrapper_list; ctx; ctx = ctx->next)
         {
           /* Check whether there is any logging to be done. */
-          if (nfds && ctx->log_fd != -1 && FD_ISSET (ctx->log_fd, &read_fds))
+          if (nfds && ctx->log_fd != -1
+	      && pth_event_status (ctx->log_ev) == PTH_STATUS_OCCURRED)
             {
-              /* Note that we do not need to use pth_read here because
-                 we already know that the read won't block. */
+              /* We must use the pth_read function for pipes, always.  */
               do 
-                n = read (ctx->log_fd, line, sizeof line - 1);
+                n = pth_read (ctx->log_fd, line, sizeof line - 1);
               while (n < 0 && errno == EINTR);
               if (n < 0)
                 {
@@ -294,8 +322,10 @@ ldap_wrapper_thread (void *dummy)
               else if (!n) /* EOF */
                 {
                   print_log_line (ctx, NULL);
-                  close (ctx->log_fd);
+                  pth_close (ctx->log_fd);
                   ctx->log_fd = -1;
+                  pth_event_free (ctx->log_ev, PTH_FREE_THIS);
+                  ctx->log_ev = NULL;
                 }
               else 
                 {
@@ -309,35 +339,31 @@ ldap_wrapper_thread (void *dummy)
           /* Check whether the process is still running.  */
           if (ctx->pid != (pid_t)(-1))
             {
-              int i, status;
+              gpg_error_t err;
+	      int status;
               
-              while ( (i=waitpid (ctx->pid, &status, WNOHANG)) == -1
-                      && errno == EINTR)
-                ;
-              if (i == -1)
-                log_error (_("waiting for ldap wrapper %d failed: %s\n"),
-                           (int)ctx->pid, strerror (errno));
-              else if (i)
+	      err = dirmngr_wait_process (ctx->pid, 0, &status);
+              if (!err)
                 {
-                  if (!WIFEXITED (status))
-                    log_info (_("ldap wrapper %d ready: terminated\n"),
-                              (int)ctx->pid);
-                  else if (WEXITSTATUS (status) == 10 )
-                    log_info (_("ldap wrapper %d ready: timeout\n"),
-                              (int)ctx->pid);
-                  else 
-                    log_info (_("ldap wrapper %d ready: exit status %d\n"),
-                              (int)ctx->pid, WEXITSTATUS (status));
+		  log_info (status == 10 ?
+                            _("ldap wrapper %d ready: timeout\n") :
+                            _("ldap wrapper %d ready"), (int)ctx->pid);
                   ctx->ready = 1;
+		  dirmngr_release_process (ctx->pid);
                   ctx->pid = (pid_t)(-1);
                 }
+              else if (gpg_err_code (err) != GPG_ERR_TIMEOUT)
+                {
+                  log_error (_("waiting for ldap wrapper %d failed: %s\n"),
+                             (int)ctx->pid, gpg_strerror (err));
+                }
             }
 
           /* Check whether we should terminate the process. */
           if (ctx->pid != (pid_t)(-1)
               && ctx->stamp != (time_t)(-1) && ctx->stamp < current_time)
             {
-              if (!kill (ctx->pid, SIGTERM))
+              if (!dirmngr_kill_process (ctx->pid))
                 ctx->stamp = (time_t)(-1);
               log_info (_("ldap wrapper %d stalled - killing\n"),
                         (int)ctx->pid);
@@ -396,7 +422,7 @@ ldap_wrapper_release_context (ksba_reader_t reader)
           {
             fd = ctx->fd;
             ctx->fd = -1;
-            close (fd);
+            pth_close (fd);
           }
         if (ctx->ctrl)
           {
@@ -423,7 +449,7 @@ ldap_wrapper_connection_cleanup (ctrl_t ctrl)
         ctx->ctrl->refcount--;
         ctx->ctrl = NULL;
         if (ctx->pid != (pid_t)(-1))
-          kill (ctx->pid, SIGTERM);
+          dirmngr_kill_process (ctx->pid);
         if (ctx->fd_error)
           log_info (_("reading from ldap wrapper %d failed: %s\n"),
                     ctx->printable_pid, gpg_strerror (ctx->fd_error));
@@ -463,16 +489,16 @@ reader_callback (void *cb_value, char *buffer, size_t count,  size_t *nread)
       pth_event_t evt;
       gpg_error_t err;
 
-      evt = pth_event (PTH_EVENT_TIME, pth_timeout (1, 0) );
-      n = pth_read_ev ( ctx->fd, buffer, nleft, evt);
-      if (n < 0 && evt && pth_event_occurred (evt) )
+      evt = pth_event (PTH_EVENT_TIME, pth_timeout (1, 0));
+      n = pth_read_ev (ctx->fd, buffer, nleft, evt);
+      if (n < 0 && evt && pth_event_occurred (evt))
         {
           n = 0;
           err = dirmngr_tick (ctx->ctrl);
           if (err)
             {
               ctx->fd_error = err;
-              close (ctx->fd);
+              pth_close (ctx->fd);
               ctx->fd = -1;
               if (evt)
                 pth_event_free (evt, PTH_FREE_THIS);
@@ -483,7 +509,7 @@ reader_callback (void *cb_value, char *buffer, size_t count,  size_t *nread)
       else if (n < 0)
         {
           ctx->fd_error = gpg_error_from_errno (errno);
-          close (ctx->fd);
+          pth_close (ctx->fd);
           ctx->fd = -1;
           if (evt)
             pth_event_free (evt, PTH_FREE_THIS);
@@ -513,9 +539,8 @@ reader_callback (void *cb_value, char *buffer, size_t count,  size_t *nread)
 }
 
 /* Fork and exec the LDAP wrapper and returns a new libksba reader
-   object at READER.  ARGV is a NULL terminated list or argumenst for
-   the wrapper; however the function adds the program's name as the
-   first arg.  The function returns 0 on success or an error code.
+   object at READER.  ARGV is a NULL terminated list or arguments for
+   the wrapper.  The function returns 0 on success or an error code.
 
    We can't use LDAP directly for these reasons:
 
@@ -529,31 +554,35 @@ reader_callback (void *cb_value, char *buffer, size_t count,  size_t *nread)
    3. There is no easy way for timeouts. In particular the timeout
       value does not work for DNS lookups (well, this is usual) and it
       seems not to work while loading a large attribute like a
-      CRL. Having a separate process allows us to either tell than
+      CRL. Having a separate process allows us to either tell the
       process to commit suicide or have our own housekepping function
       kill it after some time.  The latter also allows proper
-      cancellation of a query at any point.
+      cancellation of a query at any point of time.
       
    4. Given that we are going out to the network and usually get back
-      a long response, the frok/exec overhead is acceptable.
+      a long response, the fork/exec overhead is acceptable.
 
    Special hack to avoid passing a password through the command line
-   which is gloabbally visible: If the first element of ARGV is
-   "--pass" it will be removed and instead the environment variable
+   which is globally visible: If the first element of ARGV is "--pass"
+   it will be removed and instead the environment variable
    DIRMNGR_LDAP_PASS will be set to the next value of ARGV.  On modern
-   OSes the environment is not visible to other other user.  For those
-   old systems where it can't be avoided, we don't want to go into the
+   OSes the environment is not visible to other users.  For those old
+   systems where it can't be avoided, we don't want to go into the
    hassle of passing the password via stdin; it's just too complicated
-   and an LDAP password used for public directory lookups should be
-   that confidential.
-  */
+   and an LDAP password used for public directory lookups should not
+   be that confidential.  */
 static gpg_error_t
 ldap_wrapper (ctrl_t ctrl, ksba_reader_t *reader, const char *argv[])
 {
   gpg_error_t err;
   pid_t pid;
-  int rp[2], rp2[2];
   struct wrapper_context_s *ctx;
+  int i;
+  int j;
+  char **arg_list;
+  char *pgmname;
+  int fd_out;
+  int fd_err;
 
   /* It would be too simple to connect stderr just to our logging
      stream.  The problem is that if we are running multi-threaded
@@ -566,127 +595,55 @@ ldap_wrapper (ctrl_t ctrl, ksba_reader_t *reader, const char *argv[])
 
   *reader = NULL;
 
-  if (pipe (rp) == -1)
-    {
-      err = gpg_error_from_errno (errno);
-      log_error (_("error creating a pipe: %s\n"), strerror (errno));
-      return err;
-    }
-  if (pipe (rp2) == -1)
-    {
-      err = gpg_error_from_errno (errno);
-      log_error (_("error creating a pipe: %s\n"), strerror (errno));
-      close (rp[0]);
-      close (rp[1]);
-      return err;
-    }
-      
-  pid = pth_fork ();
-  if (pid == (pid_t)(-1))
-    {
-      err = gpg_error_from_errno (errno);
-      log_error (_("error forking process: %s\n"), strerror (errno));
-      close (rp[0]);
-      close (rp[1]);
-      close (rp2[0]);
-      close (rp2[1]);
-      return err;
-    }
+  /* Files: We need to prepare stdin and stdout.  We get stderr from
+     the function.  */
+  if (!opt.ldap_wrapper_program || !*opt.ldap_wrapper_program)
+    pgmname = (char *) get_dirmngr_ldap_path ();
+  else
+    pgmname = opt.ldap_wrapper_program;
 
-  if (!pid)
-    { /* Child. */
-      char *pgmname;
-      char **arg_list;
-      int n, i, j;
-      int fd;
+  /* Create command line argument array.  */
+  for (i = 0; argv[i]; i++)
+    ;
+  arg_list = xcalloc (i + 2, sizeof *arg_list);
+  for (i = j = 0; argv[i]; i++, j++)
+    if (!i && argv[i + 1] && !strcmp (*argv, "--pass"))
+      {
+	arg_list[j] = "--env-pass";
+	setenv ("DIRMNGR_LDAP_PASS", argv[1], 1);
+	i++;
+      }
+    else
+      arg_list[j] = (char*) argv[i];
 
-      if (!opt.ldap_wrapper_program || !*opt.ldap_wrapper_program)
-        pgmname = DIRMNGR_LIBEXECDIR "/dirmngr_ldap";
-      else
-        pgmname = opt.ldap_wrapper_program;
-
-      /* Create command line argument array.  */
-      for (i=0; argv[i]; i++)
-        ;
-      arg_list = xcalloc (i+3, sizeof *arg_list);
-      arg_list[0] = strrchr (pgmname, '/');
-      if (arg_list[0])
-        arg_list[0]++;
-      else
-        arg_list[0] = pgmname;
-      for (i=0,j=1; argv[i]; i++, j++)
-        if (!i && argv[i+1] && !strcmp (*argv, "--pass"))
-          {
-            arg_list[j] = "--env-pass";
-            setenv ("DIRMNGR_LDAP_PASS", argv[1], 1 );
-            i++;
-          }
-        else
-          arg_list[j] = (char*)argv[i];
-
-      /* Connect stdin to /dev/null.  */
-      fd = open ("/dev/null", O_RDONLY);
-      if (fd == -1)
-        {
-          log_error (_("can't open `%s': %s\n"), "/dev/null",strerror (errno));
-          _exit (4);
-        }
-      if (fd != STDIN_FILENO && dup2 (fd, STDIN_FILENO) == -1)
-        {
-          log_error (_("dup2 failed in child: %s\n"), strerror (errno));
-          _exit (4);
-        }
-
-      /* Connect stdout to the first pipe.  */
-      if (rp[1] != STDOUT_FILENO && dup2 (rp[1], STDOUT_FILENO) == -1)
-        {
-          log_error (_("dup2 failed in child: %s\n"), strerror (errno));
-          _exit (4);
-        }
-
-      /* Connect stderr to the second pipe.  */
-      if (rp2[1] != STDERR_FILENO && dup2 (rp2[1], STDERR_FILENO) == -1)
-        {
-          log_error (_("dup2 failed in child: %s\n"), strerror (errno));
-          _exit (4);
-        }
-
-  
-      /* Close all files which will not be duped. */
-      n = sysconf (_SC_OPEN_MAX);
-      if (n < 0)
-        n = MAX_OPEN_FDS;
-      for (i=0; i < n; i++)
-        {
-          if ( i == STDIN_FILENO || i == STDOUT_FILENO || i == STDERR_FILENO)
-            continue;
-          close(i);
-        }
-      errno = 0;
-
-      execv (pgmname, arg_list);
-      log_error (_("error running `%s': %s\n"), pgmname, strerror (errno));
-      _exit (31);
-    }
-
-  /* Parent. */
-  close (rp[1]);
-  close (rp2[1]);
-  
   ctx = xtrycalloc (1, sizeof *ctx);
   if (!ctx)
     {
       err = gpg_error_from_errno (errno);
       log_error (_("error allocating memory: %s\n"), strerror (errno));
-      kill (pid, SIGTERM);
-      close (rp[0]);
-      close (rp2[0]);
+      xfree (arg_list);
       return err;
     }
+
+  err = dirmngr_spawn_process (pgmname, arg_list,
+			       &fd_out, &fd_err, &pid);
+  xfree (arg_list);
+  if (err)
+    {
+      xfree (ctx);
+      return err;
+    }
+
   ctx->pid = pid;
-  ctx->printable_pid = (int)pid;
-  ctx->fd = rp[0];
-  ctx->log_fd = rp2[0];
+  ctx->printable_pid = (int) pid;
+  ctx->fd = fd_out;
+  ctx->log_fd = fd_err;
+  ctx->log_ev = pth_event (PTH_EVENT_FD | PTH_UNTIL_FD_READABLE, fd_err);
+  if (! ctx->log_ev)
+    {
+      xfree (ctx);
+      return gpg_error_from_syserror ();
+    }
   ctx->ctrl = ctrl;
   ctrl->refcount++;
   ctx->stamp = time (NULL);
@@ -712,7 +669,7 @@ ldap_wrapper (ctrl_t ctrl, ksba_reader_t *reader, const char *argv[])
     log_info (_("ldap wrapper %d started\n"), (int)ctx->pid);
 
   /* Need to wait for the first byte so we are able to detect an empty
-     output and not let the consumer se an EOF without further error
+     output and not let the consumer see an EOF without further error
      indications.  The CRL loading logic assumes that after return
      from this function, a failed search (e.g. host not found ) is
      indicated right away. */
@@ -841,7 +798,7 @@ url_fetch_ldap (ctrl_t ctrl, const char *url, const char *host, int port,
                           NULL, NULL, NULL, url,
                           reader);
 
-  /* FIXME: This option might be used for DoS attacks.  Becuase it
+  /* FIXME: This option might be used for DoS attacks.  Because it
      will enlarge the list of servers to consult without a limit and
      all LDAP queries w/o a host are will then try each host in
      turn. */
@@ -864,14 +821,18 @@ url_fetch_ldap (ctrl_t ctrl, const char *url, const char *host, int port,
      again using our default list of servers.  */
   if (err && !(opt.ldap_proxy && opt.only_ldap_proxy))
     {
-      ldap_server_t server;
+      struct ldapserver_iter iter;
       
       if (DBG_LOOKUP)
         log_debug ("no hostname in URL or query failed; "
                    "trying all default hostnames\n");
       
-      for (server = opt.ldapservers; err && server; server = server->next)
+      for (ldapserver_iter_begin (&iter, ctrl);
+	   err && ! ldapserver_iter_end_p (&iter);
+	   ldapserver_iter_next (&iter))
         {
+	  ldap_server_t server = iter.server;
+
           err = run_ldap_wrapper (ctrl,
                                   0,
                                   0,
@@ -896,15 +857,18 @@ gpg_error_t
 attr_fetch_ldap (ctrl_t ctrl,
                  const char *dn, const char *attr, ksba_reader_t *reader)
 {
-  struct ldap_server_s *server;
   gpg_error_t err = gpg_error (GPG_ERR_CONFIGURATION);
+  struct ldapserver_iter iter;
 
   *reader = NULL;
 
   /* FIXME; we might want to look at the Base SN to try matching
      servers first. */
-  for (server = opt.ldapservers; server; server = server->next)
+  for (ldapserver_iter_begin (&iter, ctrl); ! ldapserver_iter_end_p (&iter);
+       ldapserver_iter_next (&iter))
     {
+      ldap_server_t server = iter.server;
+
       err = run_ldap_wrapper (ctrl,
                               0,
                               0,
@@ -919,12 +883,7 @@ attr_fetch_ldap (ctrl_t ctrl,
   return err;
 }
 
-
-
-
 
-
-
 /* Parse PATTERN and return a new strlist to be used for the actual
    LDAP query.  Bit 0 of the flags field is set if that pattern is
    actually a base specification.  Caller must release the returned
@@ -1055,7 +1014,8 @@ make_url (char **url, const char *dn, const char *filter)
   char *u_dn, *u_filter;
   char const attrs[] = (USERCERTIFICATE ","
 /*                         USERSMIMECERTIFICATE "," */
-                        CACERTIFICATE);
+                        CACERTIFICATE ","
+                        X509CACERT );
 
   *url = NULL;
 
@@ -1094,14 +1054,14 @@ make_url (char **url, const char *dn, const char *filter)
 
 
 /* Prepare an LDAP query to return the attribute ATTR for the DN.  All
-   vonfigured default servers are queried until one responds.  This
+   configured default servers are queried until one responds.  This
    function returns an error code or 0 and a CONTEXT on success. */
 gpg_error_t
 start_default_fetch_ldap (ctrl_t ctrl, cert_fetch_context_t *context,
                           const char *dn, const char *attr)
 {
   gpg_error_t err;
-  struct ldap_server_s *server;
+  struct ldapserver_iter iter;
 
   *context = xtrycalloc (1, sizeof **context);
   if (!*context)
@@ -1110,8 +1070,12 @@ start_default_fetch_ldap (ctrl_t ctrl, cert_fetch_context_t *context,
   /* FIXME; we might want to look at the Base SN to try matching
      servers first. */
   err = gpg_error (GPG_ERR_CONFIGURATION);
-  for (server = opt.ldapservers; server; server = server->next)
+
+  for (ldapserver_iter_begin (&iter, ctrl); ! ldapserver_iter_end_p (&iter);
+       ldapserver_iter_next (&iter))
     {
+      ldap_server_t server = iter.server;
+
       err = run_ldap_wrapper (ctrl,
                               0,
                               1,
@@ -1397,6 +1361,13 @@ fetch_next_cert_ldap (cert_fetch_context_t context,
                                CACERTIFICATE);
                   okay = 1;
                 }
+              else if (!ascii_strcasecmp (p, X509CACERT))
+                {
+                  if (DBG_LOOKUP)
+                    log_debug ("fetch_next_cert_ldap: got attribute `%s'\n",
+                               CACERTIFICATE);
+                  okay = 1;
+                }
 /*               else if (!ascii_strcasecmp (p, USERSMIMECERTIFICATE)) */
 /*                 { */
 /*                   if (DBG_LOOKUP) */
