@@ -1,6 +1,6 @@
 /* dirmngr.c - LDAP access
  *	Copyright (C) 2002 Klarälvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004, 2005 g10 Code GmbH
+ *      Copyright (C) 2003, 2004, 2005, 2007, 2008 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -15,8 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -36,6 +35,7 @@
 #include "crlcache.h"
 #include "crlfetch.h"
 #include "dirmngr.h"
+#include "ldapserver.h"
 #include "ocsp.h"
 #include "certcache.h"
 #include "validate.h"
@@ -49,15 +49,6 @@
                                         gpg_error (GPG_ERR_ASS_PARAMETER), (t))
 
 
-/* Data used to associate an Assuan context with local server data */
-struct server_local_s 
-{
-  assuan_context_t assuan_ctx;
-};
-
-
-
-
 /* Copy the % and + escaped string S into the buffer D and replace the
    escape sequences.  Note, that it is sufficient to allocate the
    target string D as long as the source string S, i.e.: strlen(s)+1.
@@ -85,6 +76,78 @@ strcpy_escaped_plus (char *d, const unsigned char *s)
 }
 
 
+/* Check whether the option NAME appears in LINE */
+static int
+has_option (const char *line, const char *name)
+{
+  const char *s;
+  int n = strlen (name);
+
+  s = strstr (line, name);
+  return (s && (s == line || spacep (s-1)) && (!s[n] || spacep (s+n)));
+}
+
+/* Same as has_option but only considers options at the begin of the
+   line.  This is useful for commands which allow arbitrary strings on
+   the line.  */
+static int
+has_leading_option (const char *line, const char *name)
+{
+  const char *s;
+  int n;
+
+  if (name[0] != '-' || name[1] != '-' || !name[2] || spacep (name+2))
+    return 0;
+  n = strlen (name);
+  while ( *line == '-' && line[1] == '-' )
+    {
+      s = line;
+      while (*line && !spacep (line))
+        line++;
+      if (n == (line - s) && !strncmp (s, name, n))
+        return 1;
+      while (spacep (line))
+        line++;
+    }
+  return 0;
+}
+
+
+/* Same as has_option but does only test for the name of the option
+   and ignores an argument, i.e. with NAME being "--hash" it would
+   return a pointer for "--hash" as well as for "--hash=foo".  If
+   thhere is no such option NULL is returned.  The pointer returned
+   points right behind the option name, this may be an equal sign, Nul
+   or a space.  */
+/* static const char * */
+/* has_option_name (const char *line, const char *name) */
+/* { */
+/*   const char *s; */
+/*   int n = strlen (name); */
+
+/*   s = strstr (line, name); */
+/*   return (s && (s == line || spacep (s-1)) */
+/*           && (!s[n] || spacep (s+n) || s[n] == '=')) ? (s+n) : NULL; */
+/* } */
+
+
+/* Skip over options.  It is assumed that leading spaces have been
+   removed (this is the case for lines passed to a handler from
+   assuan).  Blanks after the options are also removed. */
+static char *
+skip_options (char *line)
+{
+  while ( *line == '-' && line[1] == '-' )
+    {
+      while (*line && !spacep (line))
+        line++;
+      while (spacep (line))
+        line++;
+    }
+  return line;
+}
+
+
 /* Common code for get_cert_local and get_issuer_cert_local. */
 static ksba_cert_t 
 do_get_cert_local (ctrl_t ctrl, const char *name, const char *command)
@@ -302,6 +365,11 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
       int i = *value? atoi (value) : 0;
       ctrl->force_crl_refresh = i;
     }
+  else if (!strcmp (key, "audit-events"))
+    {
+      int i = *value? atoi (value) : 0;
+      ctrl->audit_events = i;
+    }
   else
     return gpg_error (GPG_ERR_UNKNOWN_OPTION);
 
@@ -309,7 +377,32 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
 }
 
 
-/* IS_VALID <certificate_id>|<certificate_fpr>
+static int
+cmd_ldapserver (assuan_context_t ctx, char *line)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+  ldap_server_t server;
+  ldap_server_t *last_next_p;
+
+  while (spacep (line))
+    line++;
+  if (*line == '\0')
+    return PARM_ERROR (_("ldapserver missing"));
+
+  server = ldapserver_parse_one (line, "", 0);
+  if (! server)
+    return gpg_error (GPG_ERR_INV_ARG);
+
+  last_next_p = &ctrl->server_local->ldapservers;
+  while (*last_next_p)
+    last_next_p = &(*last_next_p)->next;
+  *last_next_p = server;
+  return 0;
+}
+
+
+/* ISVALID [--only-ocsp] [--force-default-responder]
+            <certificate_id>|<certificate_fpr>
   
    This command checks whether the certificate identified by the
    certificate_id is valid.  This is done by consulting CRLs or
@@ -317,12 +410,19 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
    are from gpg-error.h.  The command may callback using the inquire
    function.  See the manual for details.
  
-   The certificate_id is a hex encoded string consisting of two parts,
+   The CERTIFICATE_ID is a hex encoded string consisting of two parts,
    delimited by a single dot.  The first part is the SHA-1 hash of the
    issuer name and the second part the serial number.
 
    Alternatively the certificate's fingerprint may be given in which
    case an OCSP request is done before consulting the CRL.
+
+   If the option --only-ocsp is given, no fallback to a CRL check will
+   be used.
+
+   If the option --force-default-responder is given, only the default
+   OCSP responder will be used and any other methods of obtaining an
+   OCSP responder URL won't be used.
  */
 
 static int
@@ -333,7 +433,13 @@ cmd_isvalid (assuan_context_t ctx, char *line)
   gpg_error_t err;
   int did_inquire = 0;
   int ocsp_mode = 0;
+  int only_ocsp;
+  int force_default_responder;
   
+  only_ocsp = has_option (line, "--only-ocsp");
+  force_default_responder = has_option (line, "--force-default-responder");
+  line = skip_options (line);
+
   issuerhash = xstrdup (line); /* We need to work on a copy of the
                                   line because that same Assuan
                                   context may be used for an inquiry.
@@ -367,13 +473,15 @@ cmd_isvalid (assuan_context_t ctx, char *line)
       if (!opt.allow_ocsp)
         err = gpg_error (GPG_ERR_NOT_SUPPORTED);
       else
-        err = ocsp_isvalid (ctrl, NULL, NULL);
-      /* Fixme: If we got no ocsp response we should fall back to CRL
-         mode.  Thus we need to clear OCSP_MODE, get the issuerhash
-         and the serialno from the current certificate and jump to
-         again. */
+        err = ocsp_isvalid (ctrl, NULL, NULL, force_default_responder);
+      /* Fixme: If we got no ocsp response and --only-ocsp is not used
+         we should fall back to CRL mode.  Thus we need to clear
+         OCSP_MODE, get the issuerhash and the serialno from the
+         current certificate and jump to again. */
     }
-  else
+  else if (only_ocsp)
+    err = gpg_error (GPG_ERR_NO_CRL_KNOWN);
+  else 
     {
       switch (crl_cache_isvalid (ctrl,
                                  issuerhash, serialno,
@@ -523,13 +631,13 @@ cmd_checkcrl (assuan_context_t ctx, char *line)
 }
 
 
-/* CHECKOCSP [<fingerprint>]
+/* CHECKOCSP [--force-default-responder] [<fingerprint>]
 
    Check whether the certificate with FINGERPRINT (SHA-1 hash of the
    entire X.509 certificate blob) is valid or not by asking an OCSP
    responder responsible for this certificate.  The optional
    fingerprint may be used for a quick check in case an OCSP check has
-   been down for this certificate recently (we always cache OCSP
+   been done for this certificate recently (we always cache OCSP
    responses for a couple of minutes). If the fingerprint has not been
    given or there is no cached result, the function inquires the
    certificate using the
@@ -543,6 +651,10 @@ cmd_checkcrl (assuan_context_t ctx, char *line)
    its own mechanism which includes a local certificate store as well
    as a list of trusted root certifciates.
 
+   If the option --force-default-responder is given, only the default
+   OCSP responder will be used and any other methods of obtaining an
+   OCSP responder URL won't be used.
+
    The return value is the usual gpg-error code or 0 for ducesss;
    i.e. the certificate validity has been confirmed by a valid CRL.
 */
@@ -553,6 +665,10 @@ cmd_checkocsp (assuan_context_t ctx, char *line)
   gpg_error_t err;
   unsigned char fprbuffer[20], *fpr;
   ksba_cert_t cert;
+  int force_default_responder;
+  
+  force_default_responder = has_option (line, "--force-default-responder");
+  line = skip_options (line);
 
   fpr = get_fingerprint_from_line (line, fprbuffer);
   cert = fpr? get_cert_byfpr (fpr) : NULL;
@@ -590,7 +706,7 @@ cmd_checkocsp (assuan_context_t ctx, char *line)
   if (!opt.allow_ocsp)
     err = gpg_error (GPG_ERR_NOT_SUPPORTED);
   else
-    err = ocsp_isvalid (ctrl, cert, NULL);
+    err = ocsp_isvalid (ctrl, cert, NULL, force_default_responder);
 
  leave:
   if (err)
@@ -601,16 +717,71 @@ cmd_checkocsp (assuan_context_t ctx, char *line)
 
 
 
-/* LOOKUP <pattern>
-
-   Lookup certificates matching PATTERN.  To allow for multiple
-   patterns (which are ORed) quoting is required: Spaces are to be
-   translated into "+" or into "%20"; obviously this requires that the
-   usual escape quoting rules are applied.
-*/
-
 static int
-cmd_lookup (assuan_context_t ctx, char *line)
+lookup_cert_by_url (assuan_context_t ctx, const char *url)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+  gpg_error_t err = 0;
+  unsigned char *value = NULL;
+  size_t valuelen; 
+
+  /* Fetch single certificate given it's URL.  */
+  err = fetch_cert_by_url (ctrl, url, &value, &valuelen);
+  if (err)
+    {
+      log_error (_("fetch_cert_by_url failed: %s\n"), gpg_strerror (err));
+      goto leave;
+    }
+
+  /* Send the data, flush the buffer and then send an END. */
+  err = assuan_send_data (ctx, value, valuelen);      
+  if (!err)
+    err = assuan_send_data (ctx, NULL, 0);
+  if (!err)
+    err = assuan_write_line (ctx, "END");
+  if (err) 
+    {
+      log_error (_("error sending data: %s\n"), gpg_strerror (err));
+      goto leave;
+    }
+
+ leave:
+
+  return err;
+}
+
+
+/* Send the certificate, flush the buffer and then send an END. */
+static gpg_error_t
+return_one_cert (void *opaque, ksba_cert_t cert)
+{
+  assuan_context_t ctx = opaque;
+  gpg_error_t err;
+  const unsigned char *der;
+  size_t derlen;
+
+  der = ksba_cert_get_image (cert, &derlen);
+  if (!der)
+    err = gpg_error (GPG_ERR_INV_CERT_OBJ);
+  else
+    {
+      err = assuan_send_data (ctx, der, derlen);      
+      if (!err)
+        err = assuan_send_data (ctx, NULL, 0);
+      if (!err)
+        err = assuan_write_line (ctx, "END");
+    }
+  if (err) 
+    log_error (_("error sending data: %s\n"), gpg_strerror (err));
+  return err;
+}
+
+
+/* Lookup certificates from the internal cache or using the ldap
+   servers. */
+static int
+lookup_cert_by_pattern (assuan_context_t ctx, char *line, 
+                        int single, int cache_only)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err = 0;
@@ -618,9 +789,10 @@ cmd_lookup (assuan_context_t ctx, char *line)
   strlist_t sl, list = NULL;
   int truncated = 0, truncation_forced = 0;
   int count = 0;
+  int local_count = 0;
   unsigned char *value = NULL;
   size_t valuelen; 
-  ldap_server_t ldapserver;
+  struct ldapserver_iter ldapserver_iter;
   cert_fetch_context_t fetch_context;
   int any_no_data = 0;
 
@@ -647,11 +819,45 @@ cmd_lookup (assuan_context_t ctx, char *line)
         }
     }
 
-  /* Loop over all configured servers. */
-  for (ldapserver = opt.ldapservers;
-       ldapserver && ldapserver->host && !truncation_forced;
-       ldapserver = ldapserver->next)
+  /* First look through the internal cache.  The certifcates retruned
+     here are not counted towards the truncation limit.  */
+  if (single && !cache_only)
+    ; /* Do not read from the local cache in this case.  */
+  else
     {
+      for (sl=list; sl; sl = sl->next)
+        {
+          err = get_certs_bypattern (sl->d, return_one_cert, ctx);
+          if (!err)
+            local_count++;
+          if (!err && single)
+            goto ready; 
+
+          if (gpg_err_code (err) == GPG_ERR_NO_DATA)
+            {
+              err = 0;
+              if (cache_only)
+                any_no_data = 1;
+            }
+          else if (gpg_err_code (err) == GPG_ERR_INV_NAME && !cache_only)
+            {
+              /* No real fault because the internal pattern lookup
+                 can't yet cope with all types of pattern.  */
+              err = 0;
+            }
+          if (err)
+            goto ready;
+        }
+    }
+
+  /* Loop over all configured servers unless we want only the
+     certificates from the cache.  */
+  for (ldapserver_iter_begin (&ldapserver_iter, ctrl);
+       !cache_only && !ldapserver_iter_end_p (&ldapserver_iter)
+	 && ldapserver_iter.server->host && !truncation_forced;
+       ldapserver_iter_next (&ldapserver_iter))
+    {
+      ldap_server_t ldapserver = ldapserver_iter.server;
       
       if (DBG_LOOKUP)
         log_debug ("cmd_lookup: trying %s:%d base=%s\n", 
@@ -732,12 +938,15 @@ cmd_lookup (assuan_context_t ctx, char *line)
               truncation_forced = 1;
               log_info (_("max_replies %d exceeded\n"), opt.max_replies );
             }
+          if (single)
+            break;
         }
 
       end_cert_fetch (fetch_context);
     }
 
-  if (truncated || truncation_forced )
+ ready:
+  if (truncated || truncation_forced)
     {
       char str[50];
 
@@ -745,13 +954,59 @@ cmd_lookup (assuan_context_t ctx, char *line)
       assuan_write_status (ctx, "TRUNCATED", str);    
     }
 
-  if (!err && !count && any_no_data)
+  if (!err && !count && !local_count && any_no_data)
     err = gpg_error (GPG_ERR_NO_DATA);
 
  leave:
+  free_strlist (list);
+  return err;
+}
+
+
+/* LOOKUP [--url] [--single] [--cache-only] <pattern>
+
+   Lookup certificates matching PATTERN. With --url the pattern is
+   expected to be one URL.
+
+   If --url is not given:  To allow for multiple
+   patterns (which are ORed) quoting is required: Spaces are to be
+   translated into "+" or into "%20"; obviously this requires that the
+   usual escape quoting rules are applied.
+
+   If --url is given no special escaping is required because URLs are
+   already escaped this way.
+
+   if --single is given the first and only the first match will be
+   returned.  If --cache-only is _not_ given, no local query will be
+   done.
+
+   if --cache-only is given no external lookup is done so that only
+   certificates from the cache may get returned.
+*/
+
+static int
+cmd_lookup (assuan_context_t ctx, char *line)
+{
+  gpg_error_t err;
+  int lookup_url, single, cache_only;
+
+  lookup_url = has_leading_option (line, "--url");
+  single = has_leading_option (line, "--single");
+  cache_only = has_leading_option (line, "--cache-only");
+  line = skip_options (line);
+
+  if (lookup_url && cache_only)
+    err = gpg_error (GPG_ERR_NOT_FOUND);
+  else if (lookup_url && single)
+    err = gpg_error (GPG_ERR_NOT_IMPLEMENTED);
+  else if (lookup_url)
+    err = lookup_cert_by_url (ctx, line);
+  else
+    err = lookup_cert_by_pattern (ctx, line, single, cache_only);
+
   if (err)
     log_error (_("command %s failed: %s\n"), "LOOKUP", gpg_strerror (err));
-  free_strlist (list);
+
   return err;
 }
 
@@ -797,6 +1052,8 @@ cmd_listcrls (assuan_context_t ctx, char *line)
   gpg_error_t err;
   FILE *fp = assuan_get_data_fp (ctx);
 
+  (void)line;
+
   if (!fp)
     return PARM_ERROR (_("no data stream"));
 
@@ -828,6 +1085,8 @@ cmd_cachecert (assuan_context_t ctx, char *line)
   ksba_cert_t cert = NULL;
   unsigned char *value = NULL;
   size_t valuelen; 
+
+  (void)line;
       
   err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                        &value, &valuelen, MAX_CERT_LENGTH);
@@ -879,6 +1138,8 @@ cmd_validate (assuan_context_t ctx, char *line)
   ksba_cert_t cert = NULL;
   unsigned char *value = NULL;
   size_t valuelen; 
+
+  (void)line;
     
   err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                        &value, &valuelen, MAX_CERT_LENGTH);
@@ -935,6 +1196,7 @@ register_commands (assuan_context_t ctx)
     const char *name;
     int (*handler)(assuan_context_t, char *line);
   } table[] = {
+    { "LDAPSERVER", cmd_ldapserver },
     { "ISVALID",    cmd_isvalid },
     { "CHECKCRL",   cmd_checkcrl },
     { "CHECKOCSP",  cmd_checkocsp },
@@ -959,10 +1221,20 @@ register_commands (assuan_context_t ctx)
 }
 
 
+static void
+reset_notify (assuan_context_t ctx)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+
+  ldapserver_list_free (ctrl->server_local->ldapservers);
+  ctrl->server_local->ldapservers = NULL;
+}
+
+
 /* Startup the server and run the main command loop.  With FD = -1
    used stdin/stdout. */
 void
-start_command_handler (int fd)
+start_command_handler (assuan_fd_t fd)
 {
   static const char hello[] = "Dirmngr " VERSION " at your service";
   static char *hello_line;
@@ -983,7 +1255,7 @@ start_command_handler (int fd)
     
   dirmngr_init_default_ctrl (ctrl);
 
-  if (fd == -1)
+  if (fd == ASSUAN_INVALID_FD)
     {
       int filedes[2];
 
@@ -1032,12 +1304,13 @@ start_command_handler (int fd)
       hello_line[n] = 0;
     }
 
-  assuan_set_hello_line (ctx, hello_line);
-  assuan_register_option_handler (ctx, option_handler);
-
   ctrl->server_local->assuan_ctx = ctx;
   assuan_set_pointer (ctx, ctrl);
 
+  assuan_set_hello_line (ctx, hello_line);
+  assuan_register_option_handler (ctx, option_handler);
+  assuan_register_reset_notify (ctx, reset_notify);
+
   if (DBG_ASSUAN)
     assuan_set_log_stream (ctx, log_get_stream ());
 
@@ -1052,6 +1325,7 @@ start_command_handler (int fd)
           break;
         }
 
+#ifndef HAVE_W32_SYSTEM
       if (opt.verbose)
         {
           pid_t apid;
@@ -1062,6 +1336,7 @@ start_command_handler (int fd)
             log_info ("connection from process %ld (%ld:%ld)\n",
                       (long)apid, (long)auid, (long)agid);
         }
+#endif
 
       rc = assuan_process (ctx);
       if (rc)
@@ -1073,6 +1348,9 @@ start_command_handler (int fd)
   
   ldap_wrapper_connection_cleanup (ctrl);
 
+  ldapserver_list_free (ctrl->server_local->ldapservers);
+  ctrl->server_local->ldapservers = NULL;
+
   ctrl->server_local->assuan_ctx = NULL;
   assuan_deinit_server (ctx);
 
