@@ -1,6 +1,6 @@
 /* validate.c - Validate a certificate chain.
- *	Copyright (C) 2001, 2003, 2004 Free Software Foundation, Inc.
- *      Copyright (C) 2004, 2006 g10 Code GmbH
+ * Copyright (C) 2001, 2003, 2004, 2008 Free Software Foundation, Inc.
+ * Copyright (C) 2004, 2006, 2008 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -106,7 +106,7 @@ unknown_criticals (ksba_cert_t cert)
 static gpg_error_t
 check_cert_policy (ksba_cert_t cert)
 {
-  static char *allowed[] = {
+  static const char *allowed[] = {
     "2.289.9.9",
     NULL
   };
@@ -143,7 +143,7 @@ check_cert_policy (ksba_cert_t cert)
           if (p[strlen (allowed[idx])] != ':')
             continue; /* The length does not match. */
           /* Yep - it does match: Return okay. */
-          xfree (policies);
+          ksba_free (policies);
           return 0;
         }
     }
@@ -159,7 +159,7 @@ check_cert_policy (ksba_cert_t cert)
       err = gpg_error (GPG_ERR_NO_POLICY_MATCH);
     }
 
-  xfree (policies);
+  ksba_free (policies);
   return err;
 }
 
@@ -267,6 +267,74 @@ check_revocations (ctrl_t ctrl, chain_item_t chain)
 }
 
 
+/* Check whether CERT is a root certificate.  ISSUERDN and SUBJECTDN
+   are the DNs already extracted by the caller from CERT.  Returns
+   True if this is the case. */
+static int
+is_root_cert (ksba_cert_t cert, const char *issuerdn, const char *subjectdn)
+{
+  gpg_error_t err;
+  int result = 0;
+  ksba_sexp_t serialno;
+  ksba_sexp_t ak_keyid;
+  ksba_name_t ak_name;
+  ksba_sexp_t ak_sn;
+  const char *ak_name_str;
+  ksba_sexp_t subj_keyid = NULL;
+
+  if (!issuerdn || !subjectdn)
+    return 0;  /* No.  */
+
+  if (strcmp (issuerdn, subjectdn))
+    return 0;  /* No.  */
+
+  err = ksba_cert_get_auth_key_id (cert, &ak_keyid, &ak_name, &ak_sn);
+  if (err)
+    {
+      if (gpg_err_code (err) == GPG_ERR_NO_DATA)
+        return 1; /* Yes. Without a authorityKeyIdentifier this needs
+                     to be the Root certifcate (our trust anchor).  */
+      log_error ("error getting authorityKeyIdentifier: %s\n",
+                 gpg_strerror (err));
+      return 0; /* Well, it is broken anyway.  Return No. */
+    }
+
+  serialno = ksba_cert_get_serial (cert);
+  if (!serialno)
+    {
+      log_error ("error getting serialno: %s\n", gpg_strerror (err));
+      goto leave;
+    }
+
+  /* Check whether the auth name's matches the issuer name+sn.  If
+     that is the case this is a root certificate.  */
+  ak_name_str = ksba_name_enum (ak_name, 0);
+  if (ak_name_str
+      && !strcmp (ak_name_str, issuerdn) 
+      && !cmp_simple_canon_sexp (ak_sn, serialno))
+    {
+      result = 1;  /* Right, CERT is self-signed.  */
+      goto leave;
+    } 
+   
+  /* Similar for the ak_keyid. */
+  if (ak_keyid && !ksba_cert_get_subj_key_id (cert, NULL, &subj_keyid)
+      && !cmp_simple_canon_sexp (ak_keyid, subj_keyid))
+    {
+      result = 1;  /* Right, CERT is self-signed.  */
+      goto leave;
+    } 
+
+
+ leave:
+  ksba_free (subj_keyid);
+  ksba_free (ak_keyid);
+  ksba_name_release (ak_name);
+  ksba_free (ak_sn);
+  ksba_free (serialno);
+  return result; 
+}
+
 
 /* Validate the certificate CHAIN up to the trust anchor. Optionally
    return the closest expiration time in R_EXPTIME (this is useful for
@@ -358,8 +426,8 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
     {
       /* Get the subject and issuer name from the current
          certificate.  */
-      xfree (issuer);
-      xfree (subject);
+      ksba_free (issuer);
+      ksba_free (subject);
       issuer = ksba_cert_get_issuer (subject_cert, 0);
       subject = ksba_cert_get_subject (subject_cert, 0);
 
@@ -433,8 +501,9 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
         goto leave;
 
       /* Is this a self-signed certificate? */
-      if (subject && !strcmp (issuer, subject))
-        {  /* Yes. */
+      if (is_root_cert ( subject_cert, issuer, subject))
+        {  
+          /* Yes, this is our trust anchor.  */
           if (check_cert_sig (subject_cert, subject_cert) )
             {
               log_error (_("selfsigned certificate has a BAD signature"));
@@ -687,7 +756,8 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
 
   if (r_exptime)
     copy_time (r_exptime, exptime);
-  xfree (issuer);
+  ksba_free (issuer);
+  ksba_free (subject);
   ksba_cert_release (issuer_cert);
   if (subject_cert != cert)
     ksba_cert_release (subject_cert);
@@ -779,7 +849,13 @@ check_cert_sig (ksba_cert_t issuer_cert, ksba_cert_t cert)
       return err;
     }
   if (DBG_HASHING)
-    gcry_md_start_debug (md, "hash.cert");
+    {
+#ifdef HAVE_GCRY_MD_DEBUG
+      gcry_md_debug (md, "hash.cert");
+#else
+      gcry_md_start_debug (md, "hash.cert");
+#endif
+    }
   err = ksba_cert_hash (cert, 1, HASH_FNC, md);
   if (err)
     {
@@ -947,7 +1023,7 @@ cert_usage_p (ksba_cert_t cert, int mode)
               if ((p = strchr (pend, '\n')))
                 p++;
             }
-          xfree (extkeyusages);
+          ksba_free (extkeyusages);
           extkeyusages = NULL;
           
           if (!any_critical)
@@ -972,7 +1048,7 @@ cert_usage_p (ksba_cert_t cert, int mode)
     { 
       log_error (_("error getting key usage information: %s\n"),
                  gpg_strerror (err));
-      xfree (extkeyusages);
+      ksba_free (extkeyusages);
       return err;
     } 
 
